import pandas as pd
from suzieq.sqobjects.basicobj import SqObject


class PathObj(SqObject):
    '''The object providing access to the path table'''

    def __init__(self, **kwargs):
        super().__init__(table='path', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns',
                                'vrf', 'src', 'dest', 'query_str']
        self._valid_summarize_args = ['namespace', 'hostname', 'src',
                                      'dest', 'vrf', 'query_str']

    def get(self, **kwargs) -> pd.DataFrame:
        view = kwargs.get('view', self.view)
        if view == 'all':
            raise AttributeError("Cannot use 'view=all' with this table")
        return super().get(**kwargs)



from typing import List
import numpy as np
import pandas as pd

from suzieq.engines.pandas.engineobj import SqPandasEngine


class LldpObj(SqPandasEngine):
    '''Backend class to handle manipulating LLDP table with pandas'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'lldp'

    def get(self, **kwargs):
        '''A separate get to handle JunOS MACaddr as peer'''

        namespace = kwargs.get('namespace', [])
        columns = kwargs.pop('columns', [])
        use_bond = kwargs.pop('use_bond', "False")
        query_str = kwargs.pop('query_str', '')

        addnl_fields = []
        fields = self.schema.get_display_fields(columns)
        self._add_active_to_fields(kwargs.get('view', self.iobj.view), fields,
                                   addnl_fields)

        if columns == ['default']:
            needed_fields = ['subtype', 'peerMacaddr', 'peerIfindex']
        elif 'peerIfname' in columns:
            needed_fields = ['namespace', 'hostname', 'ifname', 'peerHostname',
                             'subtype', 'peerMacaddr', 'peerIfindex']
        else:
            needed_fields = []

        addnl_fields += [f for f in needed_fields if f not in fields]

        user_query_cols = self._get_user_query_cols(query_str)
        addnl_fields += [x for x in user_query_cols if x not in addnl_fields]

        df = super().get(addnl_fields=addnl_fields, columns=fields, **kwargs)
        if df.empty:
            return df
        if not needed_fields and columns != ['*']:
            return df[fields]

        macdf = df.query('subtype.isin(["", "mac address"])')
        if not macdf.empty:
            macs = macdf.peerMacaddr.unique().tolist()
            addrdf = self._get_table_sqobj('address', start_time='').get(
                namespace=namespace, address=macs,
                columns=['namespace', 'hostname', 'ifname', 'macaddr'])

            if not addrdf.empty:
                df = df.merge(
                    addrdf, how='left',
                    left_on=['namespace', 'peerHostname', 'peerMacaddr'],
                    right_on=['namespace', 'hostname', 'macaddr'],
                    suffixes=['', '_y']) \
                    .fillna({'ifname_y': '-'}) \

                if not df.empty:
                    df['peerIfname'] = np.where(df['peerIfname'] == '-',
                                                df['ifname_y'],
                                                df['peerIfname'])
                    df = df.drop(columns=['hostname_y', 'ifname_y',
                                          'timestamp_y', 'vrf', 'macaddr'],
                                 errors='ignore')

        ifidx_df = df.query('subtype.str.startswith("locally")')
        # stringify the numbers because sqobj expects pretty much all input
        # to be strings
        ifindices = [str(x)
                     for x in ifidx_df.query('peerIfindex != 0').peerIfindex
                     .unique().tolist()]
        if not ifidx_df.empty and ifindices:
            ifdf = self._get_table_sqobj('interfaces', start_time='').get(
                namespace=namespace, ifindex=ifindices,
                columns=['namespace', 'hostname', 'ifname', 'ifindex'])
            df = df.merge(
                ifdf, how='left',
                left_on=['namespace', 'peerHostname', 'peerIfindex'],
                right_on=['namespace', 'hostname', 'ifindex'],
                suffixes=['', '_y']) \
                .fillna({'ifname_y': '-'})

            if not df.empty:
                df['peerIfname'] = np.where(df['peerIfname'] == '-',
                                            df['ifname_y'], df['peerIfname'])

        # Intelligently handle hostnames without FQDN announcing themselves
        # with FQDN in LLDP
        if not df.empty:
            df = self._fix_fqdn_in_peerhost(df)

        if use_bond.lower() == "true":
            df = self._resolve_to_bond(
                df[fields], hostname=kwargs.get('hostname', []))[fields]

        df = self._handle_user_query_str(df, query_str)
        return df.reset_index(drop=True)[fields]

    def summarize(self, **kwargs):
        '''Summarize LLDP info'''
        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('nbrCnt', 'hostname', 'count'),
            ('peerHostnameCnt', 'peerHostname', 'count'),
            ('uniquePeerMgmtIPCnt', 'mgmtIP', 'nunique'),
        ]

        self._summarize_on_add_with_query = [
            ('missingPeerInfoCnt', "mgmtIP == '-' or mgmtIP == ''", 'mgmtIP')
        ]

        return super().summarize(**kwargs)

    def _resolve_to_bond(self, df: pd.DataFrame,
                         hostname=List[str]) -> pd.DataFrame:
        """Change bond/channel member ports into bond name

        With normal port channels or bond interfaces, this routine
        replaces the ifname and peerifname with the appropriate bond
        name.

        Args:
            df: The original LLDP dataframe
            hostname: The hostname selection passed by user

        Returns:
            The LLDP dataframe with the appropriate substitution
        """

        ifdf = self._get_table_sqobj('interfaces', start_time='').get(
            namespace=df.namespace.unique().tolist(),
            hostname=hostname,
            type=['bond_slave', 'bond'],
            columns=['namespace', 'hostname', 'ifname', 'adminState',
                     'ipAddressList', 'master', 'mtu'])

        if ifdf.empty:
            # Could we do something better, probably. TBD.
            return df

        # OSPF sessions maybe over a port-channel. Make LLDP
        # matchup
        lldp_df = df.merge(ifdf, on=['namespace', 'hostname', 'ifname'],
                           how='left')
        lldp_df = lldp_df.merge(
            ifdf,
            left_on=['namespace', 'peerHostname', 'peerIfname'],
            right_on=['namespace', 'hostname', 'ifname'],
            how='left', suffixes=['', '_peer']) \
            .replace(to_replace={'master': {'bridge': np.nan,
                                            'default': np.nan,
                                            '': np.nan},
                                 'master_peer': {'bridge': np.nan,
                                                 'default': np.nan,
                                                 '': np.nan}})

        lldp_df['ifname'] = np.where(~lldp_df.master.isnull(),
                                     lldp_df.master, lldp_df.ifname)
        lldp_df['peerIfname'] = np.where(~lldp_df.master_peer.isnull(),
                                         lldp_df.master_peer,
                                         lldp_df.peerIfname)
        lldp_df = lldp_df.drop_duplicates(
            subset=['namespace', 'hostname', 'ifname']) \
            .drop(columns=['master', 'master_peer'], errors='ignore')

        return lldp_df

    def _fix_fqdn_in_peerhost(self, df: pd.DataFrame) -> pd.DataFrame:
        """Intelligently strip FQDN from peer hostname if necessary.

        In some cases, the hostname returned by the NOS is without the
        domain name while the hostname advertised in LLDP is the FQDN.
        This means we cannot use LLDP to determine peering interfaces.
        This routine addresses this by intelligently stripping domain
        name from hosts if they're known without FQDN

        Args:
            df: LLDP pandas Dataframe

        Returns:
            pandas Dataframe with the peerHost entry fixed
        """
        hosts = set(df.hostname.unique().tolist())
        peerhosts = set(df.peerHostname.unique().tolist())

        ph_map = {hname: (hname.split('.')[0]
                          if hname.split('.')[0] in hosts else hname)
                  for hname in peerhosts}

        df['peerHostname'] = df.peerHostname.apply(lambda x, ph_map:
                                                   ph_map[x], args=(ph_map,))

        return df



import re

import numpy as np
from suzieq.poller.worker.services.service import Service
from suzieq.shared.utils import get_timestamp_from_junos_time, \
    parse_relative_timestamp


class DeviceService(Service):
    """Checks the uptime and OS/version of the node.
    This is specially called out to normalize the timestamp and handle
    timestamp diff
    """

    def __init__(self, name, defn, period, stype, keys, ignore_fields,
                 schema, queue, db_access, run_once):
        super().__init__(name, defn, period, stype, keys, ignore_fields,
                         schema, queue, db_access, run_once)
        self.ignore_fields.append("bootupTimestamp")

    def _common_data_cleaner(self, processed_data, raw_data):
        for entry in processed_data:
            entry['status'] = "alive"
            entry["address"] = raw_data[0]["address"]

        return processed_data

    def _clean_linux_data(self, processed_data, raw_data):

        for entry in processed_data:
            if entry.get("bootupTimestamp"):
                entry["bootupTimestamp"] = int(entry["bootupTimestamp"])

            # This is the case for Linux servers, so also extract the vendor
            # and version from the os string
            if not entry.get("vendor", ''):
                if 'os' in entry:
                    osstr = entry.get("os", "").split()
                    if len(osstr) > 1:
                        # Assumed format is: Ubuntu 18.04.2 LTS,
                        # CentOS Linux 7 (Core)
                        entry["vendor"] = osstr[0]
                        if not entry.get("version", ""):
                            entry["version"] = ' '.join(osstr[1:])
                    del entry["os"]
            entry['os'] = 'linux'

        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_cumulus_data(self, processed_data, raw_data):

        drop_indices = []
        ignore_model_info = False
        for i, entry in enumerate(processed_data):
            etype = entry.get('_entryType', '')
            if etype == 'json':
                model = entry.get('_modelName', '')
                if model:
                    entry['model'] = model

                # Some devices do not support the json output, when this is
                # available, we don't need to parse the model info as present
                # in the json output
                ignore_model_info = True
            elif etype == 'bootupTimestamp':
                entry["bootupTimestamp"] = int(entry["bootupTimestamp"])
            elif etype == 'model':
                # If we parsed the json output no need to parse this result, so
                # skip.
                if ignore_model_info:
                    drop_indices.append(i)
                else:
                    mem = entry['memory'] or 0
                    entry['mem'] = int(int(mem)/1024)
                    if not entry.get('vendor', ''):
                        entry['vendor'] = 'Cumulus'
                    entry['os'] = 'cumulus'

        processed_data = np.delete(processed_data, drop_indices).tolist()

        new_entries = {}
        for entry in processed_data:
            new_entries.update(entry)

        if new_entries:
            processed_data = [new_entries]
        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_sonic_data(self, processed_data, raw_data):
        processed_data = self._clean_linux_data(processed_data, raw_data)
        # Sonic specific updates
        for entry in processed_data:
            entry['os'] = 'sonic'
            entry['vendor'] = entry['model'].split('-')[0]

        return processed_data

    def _clean_common_ios(self, entry, os, rcv_timestamp):
        '''Common IOS-like NOS cleaning'''
        entry['os'] = os
        entry['vendor'] = 'Cisco'
        if entry.get('bootupTimestamp', ''):
            entry['bootupTimestamp'] = parse_relative_timestamp(
                entry['bootupTimestamp'], rcv_timestamp / 1000)

    def _clean_iosxr_data(self, processed_data, raw_data):
        for i, entry in enumerate(processed_data):
            self._clean_common_ios(
                entry, 'iosxr', raw_data[i]['cmd_timestamp'])
            if 'IOS-XRv' in entry.get('model', ''):
                entry['architecture'] = "x86-64"

        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_iosxe_data(self, processed_data, raw_data):
        for i, entry in enumerate(processed_data):
            self._clean_common_ios(
                entry, 'iosxe', raw_data[i]['cmd_timestamp'])

        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_ios_data(self, processed_data, raw_data):
        for i, entry in enumerate(processed_data):
            self._clean_common_ios(entry, 'ios', raw_data[i]['cmd_timestamp'])

        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_junos_data(self, processed_data, raw_data):

        for entry in processed_data:
            entry['bootupTimestamp'] = get_timestamp_from_junos_time(
                entry['bootupTimestamp'], ms=False)

            if entry.get('version', '').endswith('EVO'):
                entry['os'] = 'junos-evo'
                continue

            model = entry.get('model', '')
            if 'qfx10' in model:
                entry['os'] = 'junos-qfx10k'
            elif 'qfx' in model:
                entry['os'] = 'junos-qfx'
            elif model.startswith(('mx', 'vmx')):
                entry['os'] = 'junos-mx'
            elif 'ex' in model:
                entry['os'] = 'junos-ex'
            elif model.startswith(('srx', 'vSRX')):
                entry['os'] = 'junos-es'

        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_nxos_data(self, processed_data, raw_data):
        for entry in processed_data:
            upsecs = (24*3600*int(entry.pop('kern_uptm_days', 0)) +
                      3600*int(entry.pop('kern_uptm_hrs', 0)) +
                      60*int(entry.pop('kern_uptm_mins', 0)) +
                      int(entry.pop('kern_uptm_secs', 0)))
            if upsecs:
                entry['bootupTimestamp'] = int(
                    int(raw_data[0]['cmd_timestamp'])/1000 - upsecs)

            # Needed for textfsm parsed data
            entry['vendor'] = 'Cisco'
            entry['os'] = 'nxos'
            entry['model'] = entry.get('model', '').strip()
            entry['architecture'] = entry.get('architecture', '').strip()

        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_panos_data(self, processed_data, raw_data):
        for entry in processed_data:
            upsecs = None
            match = re.search(
                r'(\d+)\sdays,\s(\d+):(\d+):(\d+)',
                entry.get('_uptime'))
            if match:
                days = match.group(1).strip()
                hours = match.group(2).strip()
                minutes = match.group(3).strip()
                seconds = match.group(4).strip()
                upsecs = 86400 * int(days) + 3600 * int(hours) + \
                    60 * int(minutes) + int(seconds)
            if upsecs:
                entry['bootupTimestamp'] = int(
                    int(raw_data[0]["cmd_timestamp"])/1000 - upsecs)
            # defaults
            entry["vendor"] = "Palo Alto"
            entry["os"] = "panos"
        return self._common_data_cleaner(processed_data, raw_data)



from typing import List

import pandas as pd
from suzieq.sqobjects.basicobj import SqObject


class TopologyObj(SqObject):
    '''The object providing access to the virtual table: topology'''

    def __init__(self, **kwargs):
        '''Init routine'''
        super().__init__(table='topology', **kwargs)
        self._sort_fields = ["namespace", "hostname", "ifname"]
        self._cat_fields = []
        self._valid_get_args = ['namespace', 'hostname', 'columns',
                                'polled', 'ifname', 'via', 'vrf', 'asn',
                                'area', 'peerHostname', 'afiSafi', 'query_str']
        self._valid_summarize_args = ['namespace', 'hostname', 'via', 'vrf',
                                      'asn', 'area', 'query_str']
        self._valid_arg_vals = {
            'polled': ['True', 'False', 'true', 'false', '']
        }
        self._valid_summarize_args = ['namespace', 'hostname', 'via',
                                      'query_str']

    def get(self, **kwargs) -> pd.DataFrame:
        view = kwargs.get('view', self.view)
        if view == 'all':
            raise AttributeError("Cannot use 'view=all' with this table")
        return super().get(**kwargs)

    def _get_empty_cols(self, columns: List[str], fun: str, **kwargs) \
            -> List[str]:
        if fun == 'get' and columns in [['default'], ['*']]:
            def_vias = ['bgp', 'lldp', 'ospf']
            via = kwargs.get('via', []) or def_vias
            cols = self.schema.get_display_fields(columns)
            drop_cols = []
            if 'arpnd' not in via:
                drop_cols += ['arpnd', 'arpndBidir']
            if 'bgp' not in via:
                drop_cols += ['bgp', 'asn', 'peerAsn']
            if 'ospf' not in via:
                drop_cols += ['ospf', 'area']
            if 'lldp' not in via:
                drop_cols += ['lldp']

            if via == ['bgp']:
                # if only bgp is specified, the ifname column is not returned
                drop_cols += ['ifname']

            if drop_cols:
                cols = [c for c in cols if c not in drop_cols]
            return cols

        return super()._get_empty_cols(columns, fun, **kwargs)



from suzieq.engines.pandas.engineobj import SqPandasEngine


class FsObj(SqPandasEngine):
    '''Backend class to handle manipulating filesystem table with pandas'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'fs'

    def get(self, **kwargs):
        columns = kwargs.get('columns', ['default'])
        fields = self.schema.get_display_fields(columns)

        self._add_active_to_fields(
            kwargs.get('view', self.iobj.view), fields, [])

        df = super().get(**kwargs)

        if df.empty:
            return df

        return df[fields]



from typing import TypeVar, Dict, Callable, List, Optional
from abc import abstractmethod
from collections import defaultdict
import time
from datetime import datetime, timezone
import logging
import random
from http import HTTPStatus
import json
import re
import operator
from urllib.parse import urlparse
import asyncio
from asyncio.subprocess import PIPE, DEVNULL
# pylint: disable=redefined-builtin
from concurrent.futures._base import TimeoutError

from packaging import version as version_parse
import xmltodict
import asyncssh
import aiohttp

from suzieq.poller.worker.services.service import RsltToken
from suzieq.shared.utils import get_timestamp_from_junos_time, \
    known_devtypes, parse_relative_timestamp
from suzieq.shared.exceptions import (PollingError, SqPollerConfError,
                                      UnknownDevtypeError)

logger = logging.getLogger(__name__)

IOS_TIME_AFTER_DISCOVERY = 60   # time to wait after ios(xe,xr) auto-discovery

TNode = TypeVar('TNode', bound='Node')


# pylint: disable=broad-except, attribute-defined-outside-init
class Node:
    '''Class defining communicating with a device for telemetry

    The device class provides most of the basic functionality necessary to
    communicate with a device. In this base class, we don't know the device
    type i.e. whether the device is a device running EOS, IOSXE etc. Once this
    info is known, the class of the device switches from this base class to
    the device-specific class such as EosNode, IosXENode etc. Every such
    device-specific class has its own specific function for fetching the data
    that'll help determine what commands to issue to a device given its
    version, type etc.

    A device type is either determined automatically (if we use SSH transport)
    or is specified by the user (mandatory for any other transport).

    A device is using this base class then only when it doesn't know its
    device type. So, till the device type is known, we keep invoking the
    method _detect_node_type, backing off every so often. If the device is
    an unsupported device, we do mark it as "unknown" to avoid re-initing
    the info constantly.

    To reduce the overhead of setting up an SSH connection everytime we want
    to communicate, we use persistent SSH sessions to the extent possible. In
    some cases, creating this persistent session is not so easy. This is true
    for IOS devices, and so the device-specific class for those cases handle
    this in their own way. Any transport but SSH does not support persistence.

    Its possible when a device comes back up after a reboot, the device's info
    has changed such as its version. We'll need to re-init this data in order
    to continue fetching data from a device correctly in such a case. This is
    easy to do in case of SSH as the persistent connection is torn down when
    the device reboots. We don't have a good way to handle this with any other
    transport because they don;t have a persistent connection.

    The caller of this class MUST follow the following flow:
        * Create the class
        * Call initialize with the kwargs of at least the IP address of device,
          username, password (or keyfile)
        * Call run to start the node's polling
        * Call post_commands to request node to execute commands on device &
          return data. You'll need to supply a callback fn to call after the
          data is available.
    '''
    SLEEP_BET_CMDS_SLOW = 5  # seconds slow hosts wait between commands

    CONN_INITIAL_BACKOFF = 1  # (s) initial backoff time after conn failure
    CONN_INITIAL_BACKOFF_SLOW = 30  # same as before but for slow hosts
    CONNECTION_ATTEMPTS = 4  # How many connection attempts each time
    MIN_WAIT_TIME_CONN_FAIL = 60  # (s) wait time when all conn attempts failed
    MIN_WAIT_TIME_DATA_INIT_FAIL = 50  # (s) wait time dev data init failed
    # same as before but for slow hosts
    MIN_WAIT_TIME_DATA_INIT_FAIL_SLOW = 120

    # pylint: disable=too-many-statements
    async def initialize(self, **kwargs) -> TNode:
        '''Since the class is largely async, we need a diff init'''

        if not kwargs:
            raise ValueError

        self.devtype = None  # Device type
        self.pvtkey_file = ""     # SSH private keyfile
        self.prev_result = {}  # No updates if nothing changed
        self.nsname = None
        self.svc_cmd_mapping = defaultdict(lambda: {})  # Not used yet
        self.logger = logging.getLogger(__name__)
        self.port = 0
        self.backoff = 15  # secs to backoff
        self.init_again_at = 0  # after this epoch secs, try init again
        self._connect_again_at = 0  # after this epoch secs, try connection
        self.connect_timeout = kwargs.get('connect_timeout', 15)
        # _is_connecting tells if the poller is connecting, this prevents the
        # commands to proceed if the node data are not yet initialized
        self._is_connecting = False
        self.cmd_timeout = 10  # default command timeout in seconds
        self.bootupTimestamp = 0
        self.version = "all"   # OS Version to pick the right defn
        self._service_queue = None
        self._conn = None
        self._tunnel = None
        self.svcs_proc = set()
        self.error_svcs_proc = set()
        self._conn_lock = asyncio.Lock()
        self._last_exception = None
        self._exception_timestamp = None
        self._current_exception: Optional[Exception] = None
        self.api_key = None
        self._fetching_dev_data = False  # If we are fetching the dev data
        self._stdin = self._stdout = self._long_proc = None
        self._max_retries_on_auth_fail = (kwargs.get('retries_on_auth_fail')
                                          or 0) + 1
        self._retry = self._max_retries_on_auth_fail
        self._discovery_lock = asyncio.Lock()
        self._cmd_pacer = kwargs.get('cmd_pacer')
        self.per_cmd_auth = kwargs.get('per_cmd_auth', True)

        self.address = kwargs["address"]
        self.hostname = kwargs["address"]  # default till we get hostname
        self.username = kwargs.get("username", "vagrant") or "vagrant"
        self.password = kwargs.get("password", "vagrant") or "vagrant"
        self.transport = kwargs.get("transport", "ssh")
        self.nsname = kwargs.get("namespace", "default")
        self.port = kwargs.get("port", 0)
        self.devtype = None
        self.ssh_config_file = kwargs.get("ssh_config_file", None)
        self.enable_password = kwargs.get('enable_password')

        passphrase: str = kwargs.get("passphrase", None)
        jump_host = kwargs.get("jump_host", "")
        if jump_host:
            jump_result = urlparse(jump_host)
            self.jump_user = jump_result.username or self.username
            self.jump_host = jump_result.hostname
            self.jump_host_key = None
            if jump_result.port:
                self.jump_port = jump_result.port
            else:
                self.jump_port = 22
            pvtkey_file = kwargs.pop('jump_host_key_file')
            if pvtkey_file:
                self.jump_host_key = self._decrypt_pvtkey(pvtkey_file,
                                                          passphrase)
                if not self.jump_host_key:
                    raise SqPollerConfError('Unable to read private key file'
                                            f' at {pvtkey_file}')
        else:
            self.jump_host = None
            self.jump_host_key = None

        self.ignore_known_hosts = kwargs.get('ignore_known_hosts', False)
        self.slow_host = kwargs.get('slow_host', False)
        # Number of commands to issue in parallel
        if self._cmd_pacer.max_cmds:
            # Limit the num of parallel cmds we can issue when we have limits
            self.batch_size = 1
        else:
            # 4 is a number we picked to limit using up too many SSH sessions
            # Many newer implementations allow upto 5 simultaneous SSH sessions
            self.batch_size = 4
        pvtkey_file = kwargs.get("ssh_keyfile", None)
        if pvtkey_file:
            self.pvtkey = self._decrypt_pvtkey(pvtkey_file, passphrase)
            if not self.pvtkey:
                self.logger.error("ERROR: Falling back to password for "
                                  f"{self.address}:{self.port}")
                self.pvtkey = None
        else:
            self.pvtkey = None

        self._init_service_queue()

        devtype = kwargs.get("devtype", None)
        if devtype:
            self._set_devtype(devtype, '')
            self.logger.warning(
                f'{devtype} supplied for {self.address}:{self.port}')

        if self.transport == "ssh":
            self.port = self.port or 22
            # If a devtype is provided do not perform the initial connection
            # we want to do the connection once. So that we don't need to
            # open and close it in the case of 'iosxe', 'ios', 'iosxr'.
            if not self.devtype:
                await self._init_ssh(init_dev_data=False)
        elif self.transport == "https":
            self.port = self.port or 443
            if self.devtype:
                # Checking devtype to ensure we didn't get a REST transport
                # without also providing the device type.
                await self._init_rest()

        if not devtype:
            if self.transport != 'ssh':
                self.logger.error(
                    'devtype MUST be specified in inventory file if transport'
                    f'is not ssh for {self.address}')
                return self
            elif self.is_connected:
                # So we have a connection, lets figure out if we know what
                # to do with this device
                await self._detect_node_type()

        # Now we know the dev type, fetch the data we need about the
        # device to get cracking
        if not self.devtype and self._retry:
            # Unable to connect to the node, schedule later another attempt
            self._schedule_discovery_attempt()

        return self

    @property
    def last_exception(self) -> Exception:
        '''Last exception that occurred on this device'''
        return self._last_exception

    @property
    def current_exception(self) -> Optional[Exception]:
        '''The current exception faced on this device'''
        return self._current_exception

    @current_exception.setter
    def current_exception(self, val: Exception):
        '''current exception setter'''
        self._last_exception = self._current_exception
        self._current_exception = val
        if val:
            self._exception_timestamp = int(time.time()*1000)

    @property
    def is_connected(self):
        '''Is there connectivity to the device at the transport level'''
        return self._conn is not None

    def _decrypt_pvtkey(self, pvtkey_file: str, passphrase: str) -> str:
        """Decrypt private key file"""

        keydata: str = None
        if pvtkey_file:
            try:
                keydata = asyncssh.public_key.read_private_key(pvtkey_file,
                                                               passphrase)
            except Exception as e:  # pylint: disable=broad-except
                self.logger.error(
                    f"ERROR: Unable to read private key file {pvtkey_file}"
                    f"for jump host due to {e}")

        return keydata

    def _init_service_queue(self):
        if not self._service_queue:
            self._service_queue = asyncio.Queue()

    async def _close_connection(self):
        if self.is_connected:
            if self.transport == 'ssh':
                self._conn.close()
                await self._conn.wait_closed()
            elif self.transport == 'https':
                await self._conn.close()
        if self._tunnel:
            self._tunnel.close()
            await self._tunnel.wait_closed()

        self._conn = None
        self._tunnel = None
        self._stdin = self._stdout = self._stderr = None

    async def _terminate(self):
        self.logger.warning(
            f'Node: {self.hostname} received signal to terminate')
        await self._close_connection()
        return

    def _create_error(self, cmd) -> dict:
        data = {'error': str(self.current_exception)}
        if isinstance(self.current_exception, TimeoutError):
            status = HTTPStatus.REQUEST_TIMEOUT
        elif isinstance(self.current_exception, asyncssh.misc.ProtocolError):
            status = HTTPStatus.FORBIDDEN
        elif isinstance(self.current_exception,
                        asyncssh.misc.PermissionDenied):
            status = HTTPStatus.FORBIDDEN
        elif hasattr(self.current_exception, 'code'):
            status = self.current_exception.code
        else:
            status = -1
        return self._create_result(cmd, status, data)

    def _create_result(self, cmd, status, data, cmd_timestamp=None) -> dict:
        if self.port in [22, 443]:
            # Ignore port if defaults (SSH or HTTPS)
            addrstr = self.address
        else:
            addrstr = f'{self.address}:{self.port}'
        result = {
            "status": status,
            "timestamp": int(datetime.now(tz=timezone.utc).timestamp() * 1000),
            "cmd_timestamp": (int(cmd_timestamp * 1000) if cmd_timestamp
                              else None),
            "cmd": cmd,
            "devtype": self.devtype,
            "namespace": self.nsname,
            "hostname": self.hostname,
            "address": addrstr,
            "version": self.version,
            "data": data,
        }
        return result

    async def _post_result(self, service_callback: asyncio.coroutine,
                           result: List[Dict],
                           cb_token: RsltToken):
        """This function submits the result calling the service callback, this
        is a wrapper of service callback itself, allowing to perform all the
        preliminary actions before calling the callback.

        Args:
            service_callback (asyncio.coroutine): service callback
            result (List[Dict]): the result of teh command
            cb_token (RsltToken): the metadata passed between the node and the
                service.
        """
        if cb_token:
            cb_token.bootupTimestamp = self.bootupTimestamp

        await service_callback(result, cb_token)

    async def _parse_device_type_hostname(self, output, _) -> None:
        devtype = ""
        hostname = None

        if output[0]["status"] == 0:
            # don't keep trying if we're connected to an unsupported dev
            devtype = 'unsupported'
            data = output[0]["data"]
            version_str = data

            if 'Arista' in data or 'vEOS' in data:
                devtype = "eos"
            elif "JUNOS " in data:
                model = re.search(r'Model:\s+(\S+)', data)
                if model:
                    if model.group(1).startswith(('mx', 'vmx')):
                        devtype = 'junos-mx'
                    elif 'qfx10' in model.group(1):
                        devtype = 'junos-qfx10k'
                    elif 'qfx' in model.group(1):
                        devtype = 'junos-qfx'
                    elif 'ex' in model.group(1):
                        devtype = 'junos-ex'
                    elif model.group(1).startswith(('srx', 'vSRX')):
                        devtype = 'junos-es'
                if not devtype:
                    devtype = "junos"
            elif 'Junos: ' in data:
                if re.search(r'-EVO\b', data):
                    devtype = 'junos-evo'
                else:
                    devtype = 'junos'
            elif "NX-OS" in data:
                devtype = "nxos"
            elif "SONiC" in data:
                devtype = "sonic"
            elif "Cisco IOS XR" in data:
                devtype = "iosxr"
            elif any(x in data for x in ["Cisco IOS XE", "Cisco IOS-XE"]):
                devtype = "iosxe"
            elif "Cisco IOS Software" in data:
                devtype = "ios"
            else:
                self.logger.info(
                    f'{self.address}: Got unrecognized device show version: '
                    f'{data}')

            if devtype.startswith("junos"):
                hmatch = re.search(r'Hostname:\s+(\S+)\n', data)
                if hmatch:
                    hostname = hmatch.group(1)
            elif devtype == "nxos":
                hgrp = re.search(r'[Dd]evice\s+name:\s+(\S+)', data)
                if hgrp:
                    hostname = hgrp.group(1)
                else:
                    hostname = self.address
            elif devtype == "eos":
                # We'll fill in the hostname when the node gets re-init
                hostname = None
            elif devtype in ["iosxe", "ios"]:
                matchval = re.search(r'(\S+)\s+uptime', output[0]['data'])
                if matchval:
                    hostname = matchval.group(1).strip()
                else:
                    hostname = self.address
            elif devtype != "iosxr":
                hgrp = re.search(r'(\S+)\s+uptime\s+is', data)
                if hgrp:
                    hostname = hgrp.group(1)
                else:
                    hostname = self.address

        elif (len(output) > 1) and (output[1]["status"] == 0):
            devtype = 'unsupported'
            data = output[1]["data"]
            if data:
                if "Cumulus Linux" in data:
                    devtype = "cumulus"
                else:
                    devtype = "linux"

                version_str = data
                # Hostname is the last line of the output
                if len(data.strip()) > 0:
                    hostname = data.splitlines()[-1].strip()

        if devtype == 'unsupported':
            if not self.current_exception:
                self.logger.info(
                    f'Unable to determine devtype for '
                    f'{self.address}:{self.port}')
                self._set_devtype(devtype, version_str)
                self._set_hostname(self.address)
                await self._close_connection()
                self.current_exception = UnknownDevtypeError()
        elif devtype:
            self.logger.warning(
                f'Detected {devtype} for {self.address}:{self.port},'
                f' {hostname}')
            self._set_devtype(devtype, version_str)
            # We don't set the hostname here because the real hostname
            # is retrieved via a different command on some platforms
            # and can contain the FQDN. The hostname stored with a
            # record needs to be one that is also used in LLDP so that we
            # can find interface peers
            self.current_exception = None

    async def _detect_node_type(self):
        '''Figure out what type of device this is: EOS, NXOS etc.

        Its only available if we're using ssh as transport
        '''

        if self.transport in ['ssh', 'local']:
            try:
                await self._get_device_type_hostname()
            except Exception:
                self.logger.exception(f'{self.address}:{self.port}: Node '
                                      'discovery failed due to exception')
                # All the exceptions related to timeouts and authentication
                # problems are already catched inside. If we get an
                # exception here, this is unexpected and most likely something
                # went wrong with the command output parsing.
                # In this case there is not point in retrying discovery, it is
                # likely a bug.
                self._retry = 0

            if not self.devtype:
                self.logger.debug(
                    f'No devtype for {self.hostname} {self.current_exception}')
                # We were not able to do the discovery, schedule a new attempt
                self._schedule_discovery_attempt()
            else:
                # IOS* closes the connection after the initial cmds
                # are executed. So close the conn at our end too
                # avoiding the initial persistent connection failure that
                # otherwise happens
                if self.devtype in ['iosxe', 'ios', 'iosxr']:
                    await self._close_connection()
                    # In this case opening and closing a connetion too quickly
                    # might cause an authentication failure. Wait some time
                    # before proceeding with the new connection
                    await asyncio.sleep(IOS_TIME_AFTER_DISCOVERY)

                # Need to initialize the node with the device-specific
                # commands
                await self._fetch_init_dev_data()
        else:
            self.logger.error(
                f'Non-SSH transport node {self.address}:{self.port} '
                'has no devtype specified. Node will not be polled')

    async def _get_device_type_hostname(self):
        """Determine the type of device we are talking to if using ssh/local"""
        # There isn't that much of a difference in running two commands versus
        # running them one after the other as this involves an additional ssh
        # setup time. show version works on most networking boxes and
        # hostnamectl on Linux systems. That's all we support today.
        await self._exec_cmd(self._parse_device_type_hostname,
                             ["show version",
                              "cat /etc/os-release && hostname"],
                             None, 'text', only_one=True)

    def _set_devtype(self, devtype: str, version_str: str) -> None:
        """Change the class based on the device type"""

        if not devtype:
            return

        if devtype == 'unsupported':
            self.devtype = devtype
            return

        if devtype not in known_devtypes():
            self.logger.error(f'An unknown devtype {devtype} is being added.'
                              f' This will cause problems. '
                              f'Node {self.address}:{self.port}')
            raise ValueError

        if self.devtype != devtype:
            self.devtype = devtype
            if self.devtype == "cumulus":
                self.__class__ = CumulusNode
            elif self.devtype == "eos":
                self.__class__ = EosNode
            elif self.devtype == "iosxe":
                # Since we need to create one single interactive session with
                # ios and iosxe, we cannot concurrently issue commands
                self.batch_size = 1
                self.__class__ = IosXENode
            elif self.devtype == "iosxr":
                self.__class__ = IosXRNode
            elif self.devtype == "ios":
                self.batch_size = 1
                self.__class__ = IOSNode
            elif self.devtype.startswith("junos"):
                self.__class__ = JunosNode
            elif self.devtype == "nxos":
                self.__class__ = NxosNode
            elif self.devtype.startswith("sonic"):
                self.__class__ = SonicNode
            elif self.devtype == "panos":
                self.__class__ = PanosNode
            elif self.devtype == "linux":
                self.__class__ = LinuxNode

        # Now invoke the class specific NOS version extraction
        if version_str:
            self._extract_nos_version(version_str)

    def _set_hostname(self, hostname=None):
        '''Set node hostname'''
        if hostname:
            self.hostname = hostname

    async def _init_jump_host_connection(
            self,
            options: asyncssh.SSHClientConnectionOptions) -> None:
        """Initialize jump host connection if necessary

        Args:
            options (asyncssh.SSHClientConnectionOptions): non-jump host opt
        """

        if self._tunnel:
            return

        if self.jump_host_key:
            jump_host_options = asyncssh.SSHClientConnectionOptions(
                client_keys=self.jump_host_key,
                connect_timeout=self.connect_timeout,
            )

            if self.ignore_known_hosts:
                jump_host_options = asyncssh.SSHClientConnectionOptions(
                    options=jump_host_options,
                    known_hosts=None
                )
            if self.ssh_config_file:
                jump_host_options = asyncssh.SSHClientConnectionOptions(
                    options=jump_host_options,
                    config_file=[self.ssh_config_file]
                )
        else:
            jump_host_options = options

        try:
            if self.jump_host:
                self.logger.info(
                    'Using jump host: %s, with username: %s, and port: %s',
                    self.jump_host, self.jump_user, self.jump_port
                )
                self._tunnel = await asyncssh.connect(
                    self.jump_host, port=self.jump_port,
                    options=jump_host_options, username=self.jump_user)
                self.logger.info(
                    'Connection to jump host %s succeeded', self.jump_host)

        except Exception as e:  # pylint: disable=broad-except
            self.logger.error(
                f'Cannot connect to jump host: {self.jump_host} ({e})')
            self.current_exception = e
            self._conn = None
            self._tunnel = None

    def _init_ssh_options(self) -> asyncssh.SSHClientConnectionOptions:
        """Build out the asycnssh options as specified by user config

        This is a routine because its used in multiple places.
        Returns:
            asyncssh.SSHClientConnectionOptions: [description]
        """
        options = asyncssh.SSHClientConnectionOptions(
            connect_timeout=self.connect_timeout,
            username=self.username,
            agent_identities=self.pvtkey if self.pvtkey else None,
            client_keys=self.pvtkey if self.pvtkey else None,
            password=self.password if not self.pvtkey else None,
            kex_algs='+diffie-hellman-group1-sha1',  # for older boxes
            encryption_algs='+aes256-cbc',           # for older boxes
        )
        if self.ignore_known_hosts:
            options = asyncssh.SSHClientConnectionOptions(
                options=options,
                known_hosts=None,
            )
        if self.ssh_config_file:
            options = asyncssh.SSHClientConnectionOptions(
                options=options,
                config=[self.ssh_config_file],
            )

        return options

    async def _init_ssh(self, init_dev_data=True, use_lock=True) -> None:
        '''Setup a persistent SSH connection to our node. This function is a
        wrapper for `_ssh_connect()` and  handles concurrent calls and
        multiple attempts in case of failure.
        It perform at most `self.CONNECTION_ATTEMPTS` connection attempts
        blocking all the pending commands. If after all the attempts, it does
        not succeed in creating the new connection, the function return and
        schedule a time for the next set of attempts.
        All the calls before this time will fail without even trying.

        In some cases such as IOSXE/XR, this may not actually setup a
        persistent SSH session, but does the basics of setting one up.

        use_lock is a critical parameter to avoid deadlocks. If too
        many services attempt to init_ssh at the same time, we'll cause
        much heartache. So, we acquire a lock to ensure this is done only
        once.

        However, when the calling party has additional work to do, the
        caller MUST acquire the lock and pass use_lock=False. And the
        caller MUST ensure they release the lock.
        '''
        if (self._retry and (not self.is_connected or self._is_connecting)
                and time.time() > self._connect_again_at):
            someone_tried = False
            if use_lock:
                # We need to check whether someone is already attempting a
                # connection. If so wait for the result and return.
                if self._conn_lock.locked():
                    someone_tried = True
                await self._conn_lock.acquire()

            # Someone else may have already succeeded in getting the SSH conn
            # or tried without any success
            if someone_tried or self.is_connected or not self._retry:
                if use_lock:
                    self._conn_lock.release()
                return

            self._is_connecting = True
            attempts = 0
            backoff_period = (self.CONN_INITIAL_BACKOFF_SLOW if self.slow_host
                              else self.CONN_INITIAL_BACKOFF)
            try:
                while not self.is_connected:
                    prev_auth_retry = self._retry
                    await self._ssh_connect()

                    if self.is_connected:
                        self.logger.info('Connection succeeded via SSH for '
                                         f'{self.hostname}')
                        break

                    # As we want to do all the maximum authentication attempts,
                    # reduce `attempts` only if the reason is not an
                    # authentication error
                    if prev_auth_retry == self._retry:
                        attempts += 1

                    if not self._retry or attempts >= self.CONNECTION_ATTEMPTS:
                        break

                    # Wait a backoff time and retry with the connection
                    await asyncio.sleep(backoff_period)
                    backoff_period *= 2
                    backoff_period = min(backoff_period, 120)

                # We need to be sure that devtype is set, otherwise the
                # _fetch_dev_data function is not implemented and will raise.
                # Moreover we want to be connected if we call this function
                if self.is_connected:
                    self._connect_again_at = 0
                    if init_dev_data and self.devtype:
                        await self._fetch_init_dev_data(reconnect=False)
                elif self.devtype and self._retry:
                    # If we are not able to connect after all the previous
                    # attempts
                    # wait some time before the next attempts.
                    # If we don't know the devtype we want to follow the device
                    # init scheduling, so skip.
                    wait_time = max(self.MIN_WAIT_TIME_CONN_FAIL,
                                    backoff_period * 3)
                    self._connect_again_at = time.time() + wait_time
                    next_time = datetime.fromtimestamp(self._connect_again_at)
                    logger.info(
                        f'Connection to {self.address}:{self.port} will be '
                        f'retried from {next_time}'
                    )
            finally:
                self._is_connecting = False
                if use_lock:
                    self._conn_lock.release()

    async def _ssh_connect(self):
        """Create a new SSH connection with the node. The function just creates
        the new connection without handling concurrent calls or an already
        opened connection. Useful to be overridden when the node requires
        specific operations to set up a new SSH connection (i.e. IOSXE, IOSXR).
        To safely create a new connection always call `_init_ssh()`.
        """
        options = self._init_ssh_options()
        if self.jump_host and not self._tunnel:
            await self._init_jump_host_connection(options)
            if not self._tunnel:
                return

        async with self._cmd_pacer.wait():
            try:
                if self._tunnel:
                    self._conn = await self._tunnel.connect_ssh(
                        self.address, port=self.port,
                        username=self.username,
                        options=options)
                else:
                    self._conn = await asyncssh.connect(
                        self.address,
                        username=self.username,
                        port=self.port,
                        options=options)
                self.logger.info(
                    f"Connected to {self.address}:{self.port} at "
                    f"{time.time()}")
                # Reset authentication fail attempt on success
                self._retry = self._max_retries_on_auth_fail
            except Exception as e:  # pylint: disable=broad-except
                if isinstance(e, asyncssh.HostKeyNotVerifiable):
                    self.logger.error(
                        f'Unable to connect to {self.address}: {self.port}'
                        ', host key is unknown. If you do not need to '
                        ' verify the host identity, add '
                        '"ignore-known-hosts: True" in the device section '
                        'of the inventory')
                elif isinstance(e, asyncssh.misc.PermissionDenied):
                    self._retry -= 1
                    error_msg = f'Authentication failed to {self.address} '
                    if not self._retry:
                        error_msg += ('Not retrying to avoid locking out '
                                      'user. Please restart poller with '
                                      'proper authentication.')
                    self.logger.error(f'{error_msg}: {e}')

                else:
                    self.logger.error('Unable to connect to '
                                      f'{self.address}:{self.port}, {e}, '
                                      f'{type(e)}')
                self.current_exception = e
                await self._close_connection()
                self._conn = None

    async def _init_rest(self, init_dev_data=True):
        """Create rest session and initialize the device data. This function
        handles concurrency and retries.

        Args:
            init_dev_data (bool, optional): Initialize the device data after
                connection. Defaults to True.
        """
        if (self._retry and (not self.is_connected or self._is_connecting)
                and time.time() > self._connect_again_at):
            someone_tried = False
            # We need to check whether someone is already attempting a
            # connection. If so wait for the result and return.
            if self._conn_lock.locked():
                someone_tried = True
            await self._conn_lock.acquire()

            # Someone else may have already succeeded in getting the SSH conn
            # or tried without any success
            if someone_tried or self.is_connected or not self._retry:
                self._conn_lock.release()
                return

            self._is_connecting = True
            backoff_period = (self.CONN_INITIAL_BACKOFF_SLOW if self.slow_host
                              else self.CONN_INITIAL_BACKOFF)
            try:
                await self._rest_connect()

                # We need to be sure that devtype is set, otherwise the
                # _fetch_dev_data function is not implemented and will raise.
                # Moreover we want to be connected if we call this function
                if self.is_connected:
                    self._connect_again_at = 0
                    if init_dev_data:
                        await self._fetch_init_dev_data(reconnect=False)
                elif self._retry:
                    # If we are not able to create a session
                    # If we don't know the devtype we want to follow the device
                    # init scheduling, so skip.
                    wait_time = max(self.MIN_WAIT_TIME_CONN_FAIL,
                                    backoff_period * 3)
                    self._connect_again_at = time.time() + wait_time
                    next_time = datetime.fromtimestamp(self._connect_again_at)
                    logger.info(
                        f'Connection to {self.address}:{self.port} will be '
                        f'retried from {next_time}'
                    )
            finally:
                self._is_connecting = False
                self._conn_lock.release()

    @abstractmethod
    async def _rest_connect(self):
        """Creates a http session to communicate with the node. This function
        doesn't handle concurrency and already existing connections and it's
        useful to be overridden when the node requires specific operations.
        To safely create a new session always call `_init_rest()`.
        """
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    # pylint: disable=unused-argument
    async def _local_gather(self, service_callback: Callable,
                            cmd_list: List[str], cb_token: RsltToken,
                            oformat: str, timeout: int):
        """Use local command execution to execute the commands requested

        This function takes the raw list of commands to execute on the device
        locally and calls the callback function with the data returned.

        Args:
            service_callback: The callback fn to call when we have the data
                for the commands requested (or error for any given command)
            cmd_list: The list of commands to be executed on the remote device
            cb_token: The callback token to be returned to the callback fn
            oformat: The output format we're expecting the data in: text, json
                or mixed
            timeout: How long to wait for the commands to complete, in secs
            only_one: Run till the first command in the list succeeds and
                then return
        """
        result = []
        for cmd in cmd_list:
            proc = await asyncio.create_subprocess_shell(cmd, stdout=PIPE,
                                                         stderr=PIPE)

            try:
                stdout, stderr = await asyncio.wait_for(
                    proc.communicate(), timeout=self.cmd_timeout)

                if not proc.returncode:
                    d = stdout.decode('ascii', 'ignore')
                    result.append(self._create_result(
                        cmd, proc.returncode, d))
                else:
                    d = stderr('ascii', 'ignore')
                    result.append(self._create_error(cmd))

            except asyncio.TimeoutError as e:
                self.current_exception = e
                result.append(self._create_error(cmd))

        await self._post_result(service_callback, result, cb_token)

    def _schedule_discovery_attempt(self):
        """Schedule a new attempt for the node discovery
        """
        # Add an additional offset to avoid all the nodes retry all together
        offset = (random.randint(0, 1000) / 1000)
        self.backoff = min(600, self.backoff * 2) + offset
        self.init_again_at = time.time() + self.backoff

        next_time = datetime.fromtimestamp(self.init_again_at)
        logger.info(
            f'Discovery of {self.address}:{self.port} will be retried '
            f'from {next_time}'
        )

    # pylint: disable=unused-argument
    async def _ssh_gather(self, service_callback: Callable,
                          cmd_list: List[str], cb_token: RsltToken,
                          oformat: str, timeout: int, reconnect: bool,
                          only_one: bool = False):
        """Use SSH to execute the commands requested

        This function takes the raw list of commands to execute on the device
        via SSH and calls the callback function with the data returned.

        Args:
            service_callback: The callback fn to call when we have the data
                for the commands requested (or error for any given command)
            cmd_list: The list of commands to be executed on the remote device
            cb_token: The callback token to be returned to the callback fn
            oformat: The output format we're expecting the data in: text, json
                or mixed
            timeout: How long to wait for the commands to complete, in secs
            reconnect: retry connection if the node is disconnected
            only_one: Run till the first command in the list succeeds and
                then return
        """
        result = []

        if cmd_list is None:
            await self._post_result(service_callback, result, cb_token)

        if not self.is_connected or self._is_connecting:
            if reconnect:
                await self._init_ssh()
            if not self.is_connected:
                self.logger.error(
                    f'Unable to connect to node {self.address}:{self.port}'
                    f'. While executing: {cmd_list}')
                for cmd in cmd_list:
                    result.append(self._create_error(cmd))
                await self._post_result(service_callback, result, cb_token)
                return

        timeout = timeout or self.cmd_timeout
        async with self._cmd_pacer.wait(self.per_cmd_auth):
            for cmd in cmd_list:
                try:
                    if self.slow_host:
                        await asyncio.sleep(self.SLEEP_BET_CMDS_SLOW)

                    cmd_timestamp = time.time()
                    output = await asyncio.wait_for(self._conn.run(cmd),
                                                    timeout=timeout)
                    if self.current_exception:
                        self.logger.info(
                            '%s recovered from previous exception',
                            self.hostname)
                        self.current_exception = None
                    result.append(self._create_result(
                        cmd, output.exit_status, output.stdout, cmd_timestamp))
                    if (output.exit_status == 0) and only_one:
                        break
                except Exception as e:
                    self.current_exception = e
                    result.append(self._create_error(cmd))
                    if not isinstance(e, asyncio.TimeoutError):
                        self.logger.error(
                            "%s output for %s failed due to %s", cmd,
                            self.hostname, e)
                        await self._close_connection()
                    else:
                        self.logger.error(
                            "%s output for %s failed due to timeout", cmd,
                            self.hostname)

                    break

        await self._post_result(service_callback, result, cb_token)

    async def _exec_cmd(self, service_callback, cmd_list, cb_token,
                        oformat='json', timeout=None, only_one=False,
                        reconnect=True):
        '''Routine to execute a given set of commands on device

        if only_one is True, commands are executed until the first one that
        succeeds, and the rest are ignored.
        '''
        if self.transport == "ssh":
            await self._ssh_gather(service_callback, cmd_list, cb_token,
                                   oformat, timeout, reconnect, only_one)
        elif self.transport == "https":
            await self._rest_gather(service_callback, cmd_list,
                                    cb_token, oformat, timeout, reconnect)
        elif self.transport == "local":
            await self._local_gather(service_callback, cmd_list,
                                     cb_token, oformat, timeout)
        else:
            self.logger.error(
                "Unsupported transport %s for node %s",
                self.transport, self.hostname)
        return

    # pylint: disable=too-many-nested-blocks
    async def _exec_service(self, service_callback, svc_defn: dict,
                            cb_token: RsltToken):
        '''Routine that determines cmdlist to be executed for given service

        And invokes appropriate transport routine to execute the cmds and
        return the data to the service that requested it.
        '''
        result = []  # same type as gather function
        cmd = None

        # If the service provided no service definition, something really
        # wrong happend. Raise an exception to stop polling.
        if not svc_defn:
            raise PollingError(f'The {cb_token.service} service did not '
                               'provide any service definition')

        try:
            if (not self.devtype and self._retry
                    and self.init_again_at < time.time()):
                # When we issue bunch of commands multiple tasks might try to
                # perform the discovery all together, we need only one of them
                # trying to perform the discovery, all the others can fail
                if not self._discovery_lock.locked():
                    async with self._discovery_lock:
                        await self._detect_node_type()

            # As after the discovery we call the _init_dev_data() devtype might
            # be set but the devdata is still not intialized, so discovery is
            # still pending. In this case we need to fail
            if (not self.devtype
                    or (self.devtype and self._discovery_lock.locked())):
                result.append(self._create_error(svc_defn.get("service", "-")))
                return await self._post_result(
                    service_callback, result, cb_token)

            if self.devtype == 'unsupported':
                # Service code 418 means I'm a teapot in http status codes
                # in other words, I won't brew coffee because I'm a teapot
                result.append(self._create_result(
                    svc_defn, 418, "No service definition"))
                self.error_svcs_proc.add(svc_defn.get("service"))
                return await self._post_result(
                    service_callback, result, cb_token)

            self.svcs_proc.add(svc_defn.get("service"))
            use = svc_defn.get(self.hostname, None)
            if not use:
                use = svc_defn.get(self.devtype, {})
            if not use:
                if svc_defn.get("service") not in self.error_svcs_proc:
                    res = self._create_result(svc_defn,
                                              HTTPStatus.NOT_FOUND,
                                              "No service definition")
                    result.append(res)
                    self.error_svcs_proc.add(svc_defn.get("service"))
                return await self._post_result(
                    service_callback, result, cb_token)

            # TODO This kind of logic should be encoded in config and node
            # shouldn't have to know about it
            if "copy" in use:
                use = svc_defn.get(use.get("copy"))

            if use:
                if isinstance(use, list):
                    # There's more than one version here, we have to pick ours
                    for item in use:
                        if item.get('version', '') != "all":
                            os_version = item['version']
                            opdict = {'>': operator.gt, '<': operator.lt,
                                      '>=': operator.ge, '<=': operator.le,
                                      '=': operator.eq, '!=': operator.ne}
                            op = operator.eq

                            for elem, val in opdict.items():
                                if os_version.startswith(elem):
                                    os_version = os_version.replace(
                                        elem, '').strip()
                                    op = val
                                    break

                            if op(version_parse.LegacyVersion(self.version),
                                    version_parse.LegacyVersion(os_version)):
                                cmd = item.get('command', None)
                                use = item
                                break
                        else:
                            cmd = item.get("command", None)
                            use = item
                            break
                else:
                    cmd = use.get("command", None)

            if not cmd:
                result.append(self._create_result(
                    svc_defn, HTTPStatus.NOT_FOUND, "No service definition"))
                self.error_svcs_proc.add(svc_defn.get("service"))
                return await self._post_result(
                    service_callback, result, cb_token)

            oformat = use.get('format', 'json')
            if not isinstance(cmd, list):
                if use.get('textfsm'):
                    oformat = 'text'
                cmdlist = [cmd]
            else:
                # TODO: Handling format for the multiple cmd case
                cmdlist = [x.get('command', '') for x in cmd]

            await self._exec_cmd(service_callback, cmdlist, cb_token,
                                 oformat=oformat, timeout=cb_token.timeout)
        except Exception as e:
            # Here we need to catch any uncatched exception as if the node
            # does not return any answer to the service, it will never schedule
            # the service again
            logger.exception(f'{self.address}:{self.port} exception raised '
                             f'while executing the {cb_token.service} '
                             'service.')
            self.current_exception = e
            if cmd:
                if not isinstance(cmd, list):
                    result.append(self._create_error(cmd))
                else:
                    for c in cmd:
                        result.append(self._create_error(c.get('command', '')))
            else:
                result.append(self._create_error(''))

            await self._post_result(service_callback, result, cb_token)

    async def _fetch_init_dev_data(self, reconnect=True):
        """Start data fetch to initialize the class with specific device attrs

        This function is a wrapper which calls the specific implementation for
        the devtype we are polling.
        """
        # If we are already fetching the device data directly return
        if self._fetching_dev_data:
            return

        # The _fetching_dev_data allows us to tell that we are already fetching
        # the device data, so if we establish a new connection, we should not
        # call that function again
        self._fetching_dev_data = True
        try:
            await self._fetch_init_dev_data_devtype(reconnect)
        finally:
            self._fetching_dev_data = False

    @abstractmethod
    async def _fetch_init_dev_data_devtype(self, reconnect: bool):
        """Start data fetch to initialize the class with specific device attrs

        This function initiates the process of fetching critical pieces
        of info about the device such as version, os type and hostname.
        The version string is specifically used to identify which command
        needs to be executed on a device.

        This is where the list of commands specific to the device for
        extracting the said info is specified.

        Args:
            reconnect (bool): try connection if the poller is not connected to
                the device.
        """
        raise NotImplementedError(
            f'{self.address}: _fetch_init_dev_data_devtype not implmeneted in '
            'base Node class')

    async def _parse_init_dev_data(self, output: List,
                                   cb_token: RsltToken) -> None:
        """Parse the version, uptime and hostname info from the output

        This function is a wrapper which calls the specific implementation for
        the devtype we are polling.
        """
        # Check whether all the commands has been successful
        failed_commands = [out for out in output
                           if out['status'] not in [0, 200]]
        if failed_commands:
            # Get the reason for the failure of the commands
            fail_reason = ''
            for cmd in failed_commands:
                fail_reason += (f"Cmd `{cmd['cmd']}` failed with reason: "
                                f"{cmd['data']}. ")

            # Schedule a new connection attempt to retry the discovery of the
            # node
            min_wait_time = (self.MIN_WAIT_TIME_DATA_INIT_FAIL_SLOW
                             if self.slow_host
                             else self.MIN_WAIT_TIME_DATA_INIT_FAIL)
            wait_time = random.randint(min_wait_time, min_wait_time + 30)

            self._connect_again_at = time.time() + wait_time
            if self.is_connected:
                await self._close_connection()

            next_time = datetime.fromtimestamp(self._connect_again_at)
            logger.warning(
                f'{self.address}:{self.port} unable to initialize the device '
                f'data. {fail_reason} Disconnecting and scheduling '
                f'another connection attempt from {next_time}.')
        else:
            await self._parse_init_dev_data_devtype(output, cb_token)

    @abstractmethod
    async def _parse_init_dev_data_devtype(self, output: List,
                                           cb_token: RsltToken) -> None:
        """Parse the version, uptime and hostname info from the output

        This function is the callback that extracts the uptime and hostname
        from the data fetched from the device. This function calls the device
        specific version extraction function to extract the version.

        Args:
            output: The list of outputs, one per command specified
            cb_token: The callback token we passed to the data fetcher function
        """
        raise NotImplementedError(
            f'{self.address}: parsing init base Node class')

    @abstractmethod
    def _extract_nos_version(self, data: str) -> str:
        """Extract the version string from the output passed

        Args:
            data: The output of the command to search for the version
                string

        Returns:
            The version as a string
        """
        raise NotImplementedError(
            f'{self.address}: extracting NOS in init base Node class')

    @abstractmethod
    async def _rest_gather(self, service_callback: Callable,
                           cmd_list: List[str], cb_token: RsltToken,
                           oformat: str = "json", timeout: int = None,
                           reconnect=True):
        """Use HTTP(s) to execute the commands requested

        This function takes the raw list of commands to execute on the device
        via a REST API and calls the callback function with the data returned.

        Args:
            service_callback: The callback fn to call when we have the data
                for the commands requested (or error for any given command)
            cmd_list: The list of commands to be executed on the remote device
            cb_token: The callback token to be returned to the callback fn
            oformat: The output format we're expecting the data in: text, json
                or mixed
            timeout: How long to wait for the commands to complete, in secs
            only_one: Run till the first command in the list succeeds and
                then return
        """
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    def post_commands(self, service_callback: Callable, svc_defn: Dict,
                      cb_token: RsltToken):
        """Post commands to the device for servicing

        Whenever any user of a device wishes to issue a command to
        this device, they call this function with a list of commands
        to be executed, and a callback function that is invoked when
        the commands have been executed and either there's an error or
        data is returned. Multiple commands imply the result contains
        a list of responses, one for each command that was passed. The
        result itself is a dictionary which includes the command for
        which this is the result. If the user wishes to have a token
        returned to the callback function, they can pass that via the
        cb_token parameter.

        Args:
            service_callback: The callback function
            svc_defn: The service definition file from which the
                appropriate command will be extracted for the device
            cb_token: The callback token to be passed to the callback
                function
        """
        if cb_token:
            cb_token.nodeQsize = self._service_queue.qsize()
        self._service_queue.put_nowait([service_callback, svc_defn, cb_token])

    async def run(self):
        '''Main workhorse routine for Node

        Waits for services to be executed on node via service queue,
        and executes them.
        '''

        tasks = []
        try:
            while True:
                while len(tasks) < self.batch_size:

                    request = await self._service_queue.get()

                    if request:
                        callback, service_dfn, token = request
                        tasks.append(self._exec_service(
                            callback, service_dfn, token))
                        self.logger.debug(
                            f"Scheduling {token.service} for execution")
                    if self._service_queue.empty():
                        break

                if tasks:
                    completed, pending = await asyncio.wait(
                        tasks, return_when=asyncio.FIRST_COMPLETED)

                    # Check whether the completed tasks raised an
                    # uncaught exception, as at this point there might be a
                    # service not rescheduled. So, instead of keeping the
                    # poller in an inconsistent state we raise the exception.
                    # We don't expect it to be common but when it happens we
                    # want to notice it.
                    for c in completed:
                        if raised_exp := c.exception():
                            raise raised_exp

                    tasks = list(pending)
        except asyncio.CancelledError:
            await self._terminate()
            return


class EosNode(Node):
    '''EOS Node specific implementation'''

    async def _rest_connect(self):
        '''Check that connectivity and authentication works'''

        timeout = self.cmd_timeout

        auth = aiohttp.BasicAuth(self.username,
                                 password=self.password or 'vagrant')
        url = f"https://{self.address}:{self.port}/command-api"

        try:
            self._conn = aiohttp.ClientSession(
                auth=auth, timeout=self.connect_timeout,
                connector=aiohttp.TCPConnector(ssl=False))
            async with self._conn.post(url, timeout=timeout) as response:
                _ = response.status
        except Exception as e:
            self.logger.error(
                f'Unable to connect to {self.address}:{self.port}, '
                f'error: {str(e)}')

    async def _fetch_init_dev_data_devtype(self, reconnect: bool):

        if self.transport == 'https':
            cmdlist = ["show version", "show hostname"]
        else:
            cmdlist = ["show version|json", "show hostname|json"]
        await self._exec_cmd(self._parse_init_dev_data, cmdlist,
                             None, reconnect=reconnect)

    async def _ssh_gather(self, service_callback, cmd_list, cb_token, oformat,
                          timeout, reconnect, only_one=False):
        """Use SSH to execute the commands requested

        This function takes the raw list of commands to execute on the device
        via SSH and calls the callback function with the data returned.

        Args:
            service_callback: The callback fn to call when we have the data
                for the commands requested (or error for any given command)
            cmd_list: The list of commands to be executed on the remote device
            cb_token: The callback token to be returned to the callback fn
            oformat: The output format we're expecting the data in: text, json
                or mixed
            timeout: How long to wait for the commands to complete, in secs
            only_one: Run till the first command in the list succeeds and
                then return
        """
        # We need to add the JSON option for all commands that support JSON
        # output since the command provided assumes REST API
        newcmd_list = []
        for cmd in cmd_list:
            if (oformat == "json") and not cmd.endswith('json'):
                cmd += '| json'

            newcmd_list.append(cmd)

        return await super()._ssh_gather(service_callback, newcmd_list,
                                         cb_token, oformat, timeout,
                                         reconnect, only_one)

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None, reconnect=True):

        result = []
        if not cmd_list:
            return result

        if not self.is_connected or self._is_connecting:
            if reconnect:
                await self._init_rest()
            if not self.is_connected:
                self.logger.error(
                    f'Unable to connect to node {self.address}:{self.port}'
                    f'. While executing: {cmd_list}')
                for cmd in cmd_list:
                    result.append(self._create_error(cmd))
                await self._post_result(service_callback, result, cb_token)
                return

        timeout = timeout or self.cmd_timeout

        now = int(time.time() * 1000)
        data = {
            "jsonrpc": "2.0",
            "method": "runCmds",
            "id": int(now),
            "params": {"version": 1, "format": oformat, "cmds": cmd_list},
        }
        headers = {"Content-Type": "application/json"}
        if self.port:
            url = "https://{}:{}/command-api".format(self.address, self.port)
        else:
            url = "https://{}:{}/command-api".format(self.address, self.port)

        output = []
        status = 200  # status OK

        async with self._cmd_pacer.wait(self.per_cmd_auth):
            try:
                self.logger.info(
                    f'{self.address}:{self.port} exec: {cmd_list}')
                async with self._conn.post(url, json=data,
                                           timeout=timeout,
                                           headers=headers) as response:
                    status = response.status
                    if status == HTTPStatus.OK:
                        json_out = await response.json()
                        if "result" in json_out:
                            output.extend(json_out["result"])
                        else:
                            output.extend(
                                json_out["error"].get('data', []))

                        for i, cmd in enumerate(cmd_list):
                            data = (output[i] if isinstance(output, list)
                                    else output)
                            result.append(
                                self._create_result(
                                    cmd, status, data, now / 1000)
                            )
                    else:
                        for cmd in cmd_list:
                            result.append(self._create_error(cmd))
                        self.logger.error(
                            f'{self.transport}://{self.hostname}:'
                            f'{self.port}: Commands failed due to '
                            f'{response.status}')
            except Exception as e:
                self.current_exception = e
                for cmd in cmd_list:
                    result.append(self._create_error(cmd))
                self.logger.error(
                    f"{self.transport}://{self.hostname}:{self.port}: Unable "
                    f"to communicate with node due to {str(e)}")

        await self._post_result(service_callback, result, cb_token)

    async def _parse_init_dev_data_devtype(self, output, cb_token) -> None:

        if output[0]["status"] == 0 or output[0]["status"] == 200:
            if self.transport == 'ssh':
                try:
                    data = json.loads(output[0]["data"])
                except json.JSONDecodeError:
                    self.logger.error(
                        f'nodeinit: Error decoding JSON for '
                        f'{self.address}:{self.port}')
                    return
            else:
                data = output[0]["data"]
            self.bootupTimestamp = int(data["bootupTimestamp"])
            self._extract_nos_version(data)
            if not self.version:
                self.logger.error(f'nodeinit: Error getting version for '
                                  f'{self.address}: {self.port}')

        if (len(output) > 1) and (output[1]["status"] in [0, 200]):
            if self.transport == 'ssh':
                try:
                    data = json.loads(output[1]["data"])
                except json.JSONDecodeError:
                    self.logger.error(
                        f'nodeinit: Error decoding JSON for '
                        f'{self.address}:{self.port}')
                    return
            else:
                data = output[1]["data"]
            self.hostname = data["fqdn"]

    def _extract_nos_version(self, data) -> None:
        # < 4.27 or so, the cases were different.
        if isinstance(data, str):
            match = re.search(r'Software Image Version:\s+(\S+)', data,
                              re.IGNORECASE)
            if match:
                self.version = match.group(1).strip()
            else:
                self.logger.warning(
                    f'Cannot parse version from {self.address}:{self.port}')
                self.version = "all"
        else:
            self.version = data['version']


class CumulusNode(Node):
    '''Cumulus Node specific implementation'''

    async def _rest_connect(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None, reconnect=True):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _fetch_init_dev_data_devtype(self, reconnect: bool):
        """Fill in the boot time of the node by executing certain cmds"""
        await self._exec_cmd(self._parse_init_dev_data,
                             ["cat /proc/stat | grep btime | grep -o '[0-9]*'",
                              "hostname",
                              "cat /etc/os-release"], None, 'text',
                             reconnect=reconnect)

    async def _parse_init_dev_data_devtype(self, output, _) -> None:
        """Parse the uptime command output"""

        if output[0]["status"] == 0:
            self.bootupTimestamp = int(output[0]["data"])
        if (len(output) > 1) and (output[1]["status"] == 0):
            data = output[1].get("data", '')
            hostname = data.splitlines()[0].strip()
            self.hostname = hostname

        if (len(output) > 2) and (output[2]["status"] == 0):
            data = output[2].get("data", '')
            self._extract_nos_version(data)

    def _extract_nos_version(self, data: str) -> None:
        """Extract the version from the output of /etc/os-release

        Args:
            data (str): output of show version

        Returns:
            str: version
        """
        version_str = re.search(r'VERSION_ID=(\S+)', data)
        if version_str:
            self.version = version_str.group(1).strip()
        else:
            self.version = "all"
            self.logger.error(
                f'Cannot parse version from {self.address}:{self.port}')


class LinuxNode(CumulusNode):
    '''Linux server node'''

    async def _rest_connect(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None, reconnect=True):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')


class IosXRNode(Node):
    '''IOSXR Node specific implementation'''

    async def _rest_connect(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None, reconnect=True):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _fetch_init_dev_data_devtype(self, reconnect: bool):
        """Fill in the boot time of the node by executing certain cmds"""
        await self._exec_cmd(self._parse_init_dev_data,
                             ["show version", "show run hostname"],
                             None, 'text', reconnect=reconnect)

    async def _ssh_connect(self):
        """Connect to an IOSXR device, the standard `_ssh_connect()` is not
        enough as we need to start a neverending process to keep persistent
        ssh.

        IOS XR's ssh is fragile and archaic. It doesn't support sending
        multiple commands over a single SSH connection as most other devices
        do and this may not be the only one. The bug is mentioned in
        https://github.com/ronf/asyncssh/issues/241.
        """
        await super()._ssh_connect()

        # Create a persistent ssh connection
        if self.is_connected and not self._long_proc:
            try:
                self._long_proc = await self._conn.create_process(
                    'run tail -s 3600 -f /etc/version', stdout=DEVNULL,
                    stderr=DEVNULL)
                self.logger.info(
                    f'Persistent SSH present for {self.hostname}')
            except Exception as e:
                if isinstance(e, asyncssh.misc.PermissionDenied):
                    self._retry -= 1
                self.current_exception = e
                self.logger.error('Unable to create persistent SSH session'
                                  f' for {self.hostname} due to {str(e)}')
                await self._close_connection()
                self._long_proc = None

    async def _parse_init_dev_data_devtype(self, output, cb_token) -> None:
        '''Parse the version for uptime and hostname'''
        if output[0]["status"] == 0:
            data = output[0]['data']
            timestr = re.search(r'uptime is (.*)\n', data)
            if timestr:
                timestr = timestr.group(1)
                self.bootupTimestamp = parse_relative_timestamp(
                    timestr, output[0]['cmd_timestamp'] / 1000)
            else:
                self.logger.error(
                    f'Cannot parse uptime from {self.address}:{self.port}')
                self.bootupTimestamp = -1

            self._extract_nos_version(data)

        if (len(output) > 1) and (output[1]["status"] == 0):
            data = output[1]['data']
            hostname = re.search(r'hostname (\S+)', data.strip())
            if hostname:
                self.hostname = hostname.group(1)
                self.logger.error(f'set hostname of {self.address}:{self.port}'
                                  f' to {hostname.group(1)}')

    def _extract_nos_version(self, data) -> None:
        match = re.search(r'Version\s+:\s+ (\S+)', data)
        if match:
            self.version = match.group(1).strip()
        else:
            self.logger.warning(
                f'Cannot parse version from {self.address}:{self.port}')
            self.version = "all"


class IosXENode(Node):
    '''IOS-XE Node-sepcific telemetry gather implementation'''
    IOS_DEFAULT_PROMPT = ('>', '#')  # devtype specific termination sequence

    async def _rest_connect(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None, reconnect=True):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _fetch_init_dev_data_devtype(self, reconnect: bool):
        """Fill in the boot time of the node by executing certain cmds"""
        self.prompt = self.IOS_DEFAULT_PROMPT
        await self._exec_cmd(self._parse_init_dev_data,
                             ["show version"], None, 'text',
                             reconnect=reconnect)

    async def _ssh_connect(self):
        """Connect to an IOSXR device, the standard `_ssh_connect()` is not
        enough as we need to start an interactive session for XE.
        """
        await super()._ssh_connect()

        if self.is_connected and not self._stdin:
            self.logger.info(
                f'Trying to create Persistent SSH for {self.hostname}')
            async with self._cmd_pacer.wait(self.per_cmd_auth):
                try:
                    self._stdin, self._stdout, self._stderr = \
                        await self._conn.open_session(term_type='xterm')
                    self.logger.info(
                        f'Persistent SSH created for {self.hostname}')

                    output = await self.wait_for_prompt()
                    if 'suzieq timeout' in output:
                        raise TimeoutError(
                            'Did not receive > or # prompt before privilege '
                            'escalation')

                    if output.strip().endswith('>'):
                        if await self._handle_privilege_escalation() == -1:
                            self.logger.error(f'{self.address}:{self.port}: '
                                              'Privilege escalation failed')
                            await self._close_connection()
                            self._conn = None
                            self._stdin = None
                            self._retry -= 1
                            return

                    # Set the terminal length to 0 to avoid paging
                    self.logger.info(
                        f'{self.hostname} set terminal length to 0')
                    self._stdin.write('terminal length 0\n')
                    try:
                        output = await asyncio.wait_for(
                            self._stdout.readuntil(self.prompt),
                            90)
                    except TimeoutError:
                        raise TimeoutError(
                            'Unable to set the terminal length to 0, did not '
                            'receive prompt.')
                    self.logger.info(
                        f'{self.hostname} set terminal length to 0: DONE')
                    # Reset number of retries on successful auth
                    self._retry = self._max_retries_on_auth_fail
                except Exception as e:
                    if isinstance(e, asyncssh.misc.PermissionDenied):
                        self._retry -= 1
                    self.current_exception = e
                    self.logger.error('Unable to create persistent SSH session'
                                      f' for {self.hostname} due to {str(e)}')
                    await self._close_connection()
                    return

    async def _handle_privilege_escalation(self) -> int:
        '''Escalata privilege if necessary

        Returns 0 on success, -1 otherwise'''

        self.logger.info(
            f'Privilege escalation required for {self.hostname}')
        self._stdin.write('enable\n')
        output = await self.wait_for_prompt(('Password:', '%'))
        if '%' in output:
            self.logger.error('Privilege escalation failed, Aborting')
            self.current_exception = PollingError(
                'Privilege Escalation Failed')
            return -1

        if self.enable_password:
            self._stdin.write(self.enable_password + '\n')
        else:
            self._stdin.write(self.password + '\n')

        output = await self.wait_for_prompt()
        if (output in ['suzieq timeout', 'Password:'] or
                output.strip().endswith('>')):
            self.logger.error(
                f'{self.address}:{self.port} Privilege escalation failed, '
                'Aborting connection')
            return -1

        self.logger.info(f'Privilege escalation succeeded for {self.hostname}')
        return 0

    async def wait_for_prompt(self, prompt: str = None,
                              timeout: int = 90) -> str:
        """Wait for specified prompt upto timeout duration

        Since we're waiting for a prompt, we want to not wait forever.
        asyncssh's readuntil doesn't take a timeout parameter as of
        2.9.0. So, instead of adding a asyncio.waitfor everywhere, we
        just call this routine. By default, we wait for 90s

        Args:
            prompt[str]: The prompt string to wait for
            timeout[int]: How long to wait in secs
        Returns:
            the output data or 'timeout'
        """
        if prompt is None:
            prompt = self.prompt
        data = []
        while True:
            try:
                output = await asyncio.wait_for(
                    self._stdout.readuntil(prompt), timeout=timeout)
                data.append(output)
                return ''.join(data)
            except asyncio.IncompleteReadError as exc:
                if exc.partial:
                    data.append(exc.partial)
                    self.logger.warning('Received partial data, continuing')
                else:
                    continue
            except asyncio.TimeoutError:
                self.current_exception = asyncio.TimeoutError
                self.logger.error(f'{self.address}.{self.port} '
                                  'Timed out waiting for expected prompt')
                # Return something that won't ever be in real output
                await self._close_connection()
                self.prompt = self.IOS_DEFAULT_PROMPT
                return 'suzieq timeout'

    async def _parse_init_dev_data_devtype(self, output, cb_token) -> None:
        '''Parse the version for uptime and hostname'''
        if not isinstance(output, list):
            # In some errors, the output returned is not a list
            self.bootupTimestamp = -1
            return

        if output[0]["status"] == 0:
            data = output[0]['data']
            hostupstr = re.search(r'(\S+)\s+uptime is (.*)\n', data)
            if hostupstr:
                self._set_hostname(hostupstr.group(1))
                self.prompt = tuple(f'{self.hostname}{x}'
                                    for x in self.IOS_DEFAULT_PROMPT)

                timestr = hostupstr.group(2)
                self.bootupTimestamp = parse_relative_timestamp(
                    timestr, output[0]['cmd_timestamp'] / 1000)
            else:
                self.logger.error(
                    f'Cannot parse uptime from {self.address}:{self.port}')
                self.bootupTimestamp = -1

            self._extract_nos_version(data)

    async def _ssh_gather(self, service_callback, cmd_list, cb_token, oformat,
                          timeout, reconnect, only_one=False):
        """Run ssh for cmd in cmdlist and place output on service callback
           This is different from IOSXE to avoid reinit node info each time
        """

        result = []
        if cmd_list is None:
            await self._post_result(service_callback, result, cb_token)
            return

        self.logger.info(
            f'{self.hostname} {self.address}:{self.port} '
            f'executing cmds: {cmd_list}'
        )

        if not self.is_connected or not self._stdin:
            if reconnect:
                await self._init_ssh()
            else:
                logger.debug(f'{self.address}:{self.port} is connecting, '
                             'avoid a nested connection attempt.')

        if not self._conn or not self._stdin:
            for cmd in cmd_list:
                self.logger.error(
                    "Unable to connect to node %s (%s) cmd %s",
                    self.address, self.hostname, cmd)
                result.append(self._create_error(cmd))
            await self._post_result(service_callback, result, cb_token)
            return

        timeout = timeout or self.cmd_timeout
        async with self._cmd_pacer.wait(self.per_cmd_auth):
            for cmd in cmd_list:
                try:
                    if self.slow_host:
                        await asyncio.sleep(self.SLEEP_BET_CMDS_SLOW)

                    cmd_timestamp = time.time()
                    self._stdin.write(cmd + '\n')
                    output = await self.wait_for_prompt(timeout=timeout)
                    if 'Invalid input detected' in output:
                        status = -1
                    elif 'suzieq timeout' in output:
                        status = HTTPStatus.REQUEST_TIMEOUT
                    else:
                        status = 0
                        # Strip out the command prompt that gets added
                        if output[-2:].strip() in ["#", "$"]:
                            output = output[:-2]
                    result.append(self._create_result(
                        cmd, status, output, cmd_timestamp))
                    continue
                except Exception as e:
                    self.current_exception = e
                    result.append(self._create_error(cmd))
                    if not isinstance(e, asyncio.TimeoutError):
                        self.logger.error(
                            f'{self.hostname} {self.address}:{self.port}'
                            f'output failed for {cmd} due to {e}')
                        try:
                            await self._close_connection()
                            self.logger.debug("Closed conn successfully for "
                                              f"{self.hostname}")
                        except Exception as close_exc:
                            self.logger.error(
                                f"Caught an exception closing {self.hostname}"
                                f" for {cmd}: {close_exc}")
                    else:
                        self.logger.error(
                            f'{self.hostname} {self.address}:{self.port}'
                            f'output failed for {cmd} due to timeout')
                    break

        await self._post_result(service_callback, result, cb_token)

    def _extract_nos_version(self, data: str) -> None:
        match = re.search(r', Version\s+([^ ,]+)', data)
        if match:
            self.version = match.group(1).strip()
        else:
            self.logger.warning(
                f'Cannot parse version from {self.address}:{self.port}')
            self.version = "all"


class IOSNode(IosXENode):
    '''Classic IOS Node-specific implementation'''

    async def _rest_connect(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None, reconnect=True):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')


class JunosNode(Node):
    '''Juniper's Junos node-specific implementation'''

    async def _rest_connect(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None, reconnect=True):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _fetch_init_dev_data_devtype(self, reconnect: bool):
        """Fill in the boot time of the node by running requisite cmd"""
        if self.devtype == 'junos-evo':
            cmdlist = ["show system uptime|display json",
                       'show version node all']
        else:
            cmdlist = ["show system uptime|display json",
                       'show version']
        await self._exec_cmd(self._parse_init_dev_data, cmdlist, None,
                             "mixed", reconnect=reconnect)

    async def _parse_init_dev_data_devtype(self, output, cb_token) -> None:
        """Parse the uptime command output"""
        if output[0]["status"] == 0:
            data = output[0]["data"]
            try:
                jdata = json.loads(data.replace('\n', '').strip())
                if self.devtype not in ["junos-mx", "junos-qfx10k",
                                        "junos-evo"]:
                    jdata = (jdata['multi-routing-engine-results'][0]
                             ['multi-routing-engine-item'][0])

                timestr = (jdata['system-uptime-information'][0]
                           ['system-booted-time'][0]['date-time'][0]
                           ['attributes'])
            except Exception:
                self.logger.warning(
                    f'{self.address}:{self.port} Unable to parse junos boot '
                    f'time from {data}')
                timestr = '{"junos:seconds": "0"}'
            self.bootupTimestamp = get_timestamp_from_junos_time(
                timestr, ms=False)

        if (len(output) > 1) and (output[1]["status"] == 0):
            data = output[1]["data"]
            hmatch = re.search(r'\bHostname:\s+(\S+)\b', data)
            if hmatch:
                # TODO
                # In case of a VC, we'll need to strip off some chars
                # at the end such as -re0 or -re1. Lets do this with a
                # separate option passed via the inventory device setting.
                self._set_hostname(hmatch.group(1))

            self._extract_nos_version(data)

    def _extract_nos_version(self, data) -> None:
        """Extract the version from the output of show version"""
        match = re.search(r'Junos: (\S+)', data)
        if match:
            self.version = match.group(1).strip()
        else:
            self.logger.warning(
                f'Cannot parse version from {self.address}:{self.port}')
            self.version = "all"


class NxosNode(Node):
    '''Cisco's NXOS Node-specific implementation'''

    async def _rest_connect(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None, reconnect=True):
        '''Gather data for service via device REST API'''
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _fetch_init_dev_data_devtype(self, reconnect: bool):
        """Fill in the boot time of the node by running requisite cmd"""
        await self._exec_cmd(self._parse_init_dev_data,
                             ["show version", "show hostname"], None,
                             'mixed', reconnect=reconnect)

    async def _parse_init_dev_data_devtype(self, output, cb_token) -> None:
        """Parse the uptime command output"""

        hostname = ''

        if output[0]["status"] == 0:
            data = output[0]["data"]

            self._extract_nos_version(data)
            uptime_grp = re.search(r'Kernel\s+uptime\s+is\s+([^\n]+)', data)
            if uptime_grp:
                self.bootupTimestamp = parse_relative_timestamp(
                    uptime_grp.group(1), output[0]['cmd_timestamp'] / 1000)

        if len(output) > 1:
            if output[1]["status"] == 0:
                hostname = output[1]["data"].strip()
        else:
            if output[0]['hostname'] != output[0]['address']:
                hostname = output[0]['hostname']

        if hostname:
            self._set_hostname(hostname)

    def _extract_nos_version(self, data: str) -> None:

        version = ''
        vgrp = re.search(r'system:\s+version\s+([^\n]+)', data)
        if vgrp:
            version = vgrp.group(1)
        else:
            vgrp = re.search(r'NXOS:\s+version\s+(\S+)', data)
            if vgrp:
                version = vgrp.group(1)
        if not version:
            self.logger.warning(
                f'Cannot parse version from {self.address}:{self.port}')
            self.version = "all"
        else:
            self.version = version


class SonicNode(Node):
    '''SONiC Node-specific implementtaion'''

    async def _rest_connect(self):
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None, reconnect=True):
        '''Gather data for service via device REST API'''
        raise NotImplementedError(
            f'{self.address}: REST transport is not supported')

    async def _fetch_init_dev_data_devtype(self, reconnect: bool):
        """Fill in the boot time of the node by running requisite cmd"""
        await self._exec_cmd(self._parse_init_dev_data,
                             ["cat /proc/stat | grep btime | grep -o '[0-9]*'",
                              "hostname", "show version"],
                             None, 'text', reconnect=reconnect)

    async def _parse_init_dev_data_devtype(self, output, cb_token) -> None:
        """Parse the uptime command output"""

        if output[0]["status"] == 0:
            self.bootupTimestamp = int(output[0]["data"])
        if (len(output) > 1) and (output[1]["status"] == 0):
            self.hostname = output[1]["data"].strip()
        if (len(output) > 2) and (output[2]["status"] == 0):
            self._extract_nos_version(output[1]["data"])

    def _extract_nos_version(self, data: str) -> None:
        match = re.search(r'Version:\s+SONiC-OS-([^-]+)', data)
        if match:
            self.version = match.group(1).strip()
        else:
            self.logger.warning(
                f'Cannot parse version from {self.address}:{self.port}')
            self.version = "all"


class PanosNode(Node):
    '''Node object representing access to a Palo Alto Networks FW'''

    async def _fetch_init_dev_data_devtype(self, _reconnect: bool):
        discovery_cmd = 'show system info'
        result = []
        try:
            # temporary hack to detect device info using ssh
            async with self._cmd_pacer.wait():
                async with asyncssh.connect(
                        self.address, port=22, username=self.username,
                        password=self.password, known_hosts=None) as conn:
                    async with conn.create_process() as process:
                        cmd_timestamp = time.time()
                        process.stdin.write(f'{discovery_cmd}\n')
                        output = ""
                        output += await process.stdout.read(1)
                        try:
                            await asyncio.wait_for(
                                process.wait_closed(), timeout=0.1)
                        except asyncio.TimeoutError:
                            pass

                        stdout, _ = process.collect_output()
                        output += stdout
                        result = [self._create_result(discovery_cmd, 0, output,
                                                      cmd_timestamp)]
        except Exception as e:
            self.current_exception = e
            result.append(self._create_error(discovery_cmd))
            if isinstance(e, asyncssh.misc.PermissionDenied):
                self.logger.error(
                    f'{self.address}:{self.port}: permission denied')
                self._retry -= 1
            else:
                self.logger.error(
                    f'{self.hostname}:{self.port}: Command "{discovery_cmd}" '
                    f'failed due to {e}')
        await self._parse_init_dev_data(result, None)

    async def get_api_key(self):
        """Authenticate to get the api key needed in all cmd requests"""
        url = f"https://{self.address}:{self.port}/api/?type=keygen&user=" \
            f"{self.username}&password={self.password}"

        if not self._retry:
            return
        async with self._cmd_pacer.wait(self.per_cmd_auth):
            async with self._conn.get(url, timeout=self.connect_timeout) \
                    as response:
                status, xml = response.status, await response.text()
                if status == 200:
                    data = xmltodict.parse(xml)
                    self.api_key = data["response"]["result"]["key"]
                    # reset retry count, just in case.
                    self._retry = self._max_retries_on_auth_fail
                elif status == 403:
                    self.logger.error('Invalid credentials, could not get api '
                                      f'key for {self.address}:{self.port}.')
                    self._retry -= 1
                else:
                    self.logger.error('Unknown error, could not get '
                                      'api key for '
                                      f'{self.address}:{self.port}.')

    async def _parse_init_dev_data_devtype(self, output, cb_token) -> None:
        """Parse the uptime command output"""
        if output[0]["status"] == 0:
            data = output[0]["data"]
            # extract uptime
            match = re.search(
                r'uptime:\s+(\d+)\sdays,\s(\d+):(\d+):(\d+)', data)

            if match:
                days = match.group(1).strip()
                hours = match.group(2).strip()
                minutes = match.group(3).strip()
                seconds = match.group(4).strip()
                upsecs = 86400 * int(days) + 3600 * int(hours) + \
                    60 * int(minutes) + int(seconds)
                self.bootupTimestamp = int(
                    (output[0]['cmd_timestamp'] / 1000) - float(upsecs))
            else:
                self.logger.warning(
                    f'Cannot parse uptime from {self.address}:{self.port}')

            # extract hostname
            hmatch = re.search(r'hostname:\s+(\S+)\n', data)
            if hmatch:
                self.hostname = hmatch.group(1).strip()
            else:
                self.logger.warning(
                    f'Cannot parse hostname from {self.address}:{self.port}')

            self._extract_nos_version(data)

    def _extract_nos_version(self, data: str) -> None:
        vmatch = re.search(r'sw-version:\s+(\S+)', data)
        if vmatch:
            self.version = vmatch.group(1).strip()
        else:
            self.logger.warning(
                f'Cannot parse version from {self.address}:{self.port}')
            self.version = "all"

    async def _rest_connect(self):
        # In case of PANOS, getting here means REST is up
        if not self._conn:
            async with self._cmd_pacer.wait(self.per_cmd_auth):
                try:
                    self._conn = aiohttp.ClientSession(
                        conn_timeout=self.connect_timeout,
                        connector=aiohttp.TCPConnector(ssl=False),
                    )
                    if self.api_key is None:
                        await self.get_api_key()
                    # If the api_key is still None we can't gather any data.
                    # Ensure that the connection pool is closed and set it to
                    # None so that _rest_gather can fail gracefully.
                    if self.api_key is None:
                        await self._close_connection()
                except Exception as e:
                    await self._close_connection()
                    self.logger.error(
                        f'{self.transport}://{self.hostname}:{self.port}, '
                        f'Unable to communicate due to error: {str(e)}')

    async def _rest_gather(self, service_callback, cmd_list, cb_token,
                           oformat="json", timeout=None, reconnect=True):

        result = []
        if not cmd_list:
            return result

        timeout = timeout or self.connect_timeout

        url = f"https://{self.address}:{self.port}/api/"

        status = 200  # status OK

        if not self.is_connected or self._is_connecting:
            if reconnect:
                await self._init_rest()
            if not self.is_connected:
                self.logger.error(
                    f'Unable to connect to node {self.address}:{self.port}'
                    f'. While executing: {cmd_list}')
                for cmd in cmd_list:
                    result.append(self._create_error(cmd))
                await self._post_result(service_callback, result, cb_token)
                return

        async with self._cmd_pacer.wait(self.per_cmd_auth):
            try:
                for cmd in cmd_list:
                    self.logger.info(f'{self.address}:{self.port} exec: {cmd}')
                    url_cmd = f"{url}?type=op&cmd={cmd}&key={self.api_key}"
                    cmd_timestamp = time.time()
                    async with self._conn.get(
                            url_cmd, timeout=timeout) as response:
                        status, xml = response.status, await response.text()
                        if status == 200:
                            json_out = json.dumps(
                                xmltodict.parse(xml))
                            result.append(
                                self._create_result(
                                    cmd, status, json_out, cmd_timestamp)
                            )
                        else:
                            result.append(self._create_error(cmd))
                            self.logger.error(
                                f'{self.transport}://{self.hostname}:'
                                f'{self.port}: Command {cmd} failed with '
                                f'status {response.status}')
            except Exception as e:
                self.current_exception = e
                for cmd in cmd_list:
                    result.append(self._create_error(cmd))
                self.logger.error(
                    f"{self.transport}://{self.hostname}:{self.port} "
                    f"Unable to communicate due to {str(e)}")

        await self._post_result(service_callback, result, cb_token)



import logging
from pathlib import Path


logger = logging.getLogger(__name__)


def move_broken_file(broken_file: str):
    """Move a broken file to a directory to be investigated later

    Args:
        broken_file (str): the path to the broken file.
    """
    broken_file_path = Path(broken_file)
    if broken_file_path.exists():
        # Get the path to the broken file
        splitted_path = broken_file.split('sqvers=')
        if len(splitted_path) == 2:
            table_dir = Path(splitted_path[0])
            outfile = table_dir.parent / '_broken' / table_dir.stem / \
                f"sqvers={splitted_path[1]}"

            logger.warning(
                f'Moving broken file from {broken_file} to {outfile}')
            outfile.parent.mkdir(exist_ok=True, parents=True)
            broken_file_path.rename(outfile)
        else:
            logger.error(
                f'Unable to move broken file {broken_file} wrong file path')




import time
import ast
import inspect
from io import StringIO
import shutil
from typing import List

import numpy as np
import pandas as pd
from nubia import command, context


from prompt_toolkit import prompt
from natsort import natsort_keygen
from colorama import Fore, Style

from termcolor import cprint, colored
from suzieq.cli.nubia_patch import argument
from suzieq.shared.sq_plugin import SqPlugin
from suzieq.shared.exceptions import UserQueryError
from suzieq.shared.utils import (DATA_FORMATS, SUPPORTED_ENGINES,
                                 deprecated_command_warning)


def colorize(x, color):
    '''Colorize string with the color provided'''
    return (f'{color}{Style.BRIGHT}{x}{Style.RESET_ALL}'
            if color else x)


@argument(
    "engine",
    description="Which analytical engine to use",
    choices=SUPPORTED_ENGINES,
)
@argument(
    "namespace", description="Namespace(s), space separated"
)
@argument("hostname",
          description="Hostname(s), space separated")
@argument(
    "start_time", description="Start of time window, try natural language spec"
)
@argument(
    "end_time", description="End of time window, try natural language spec "
)
@argument(
    "view",
    description="View all records or just the latest",
    choices=["all", "latest"],
)
@argument("columns", description="Space separated list of columns, * for all")
@argument(
    "format",
    description="Select the pformat of the output",
    choices=DATA_FORMATS,
)
@argument(
    "query_str",
    description="Trailing blank terminated pandas query format to "
    "further filter the output"
)
class SqCommand(SqPlugin):
    """Base Command Class for SuzieQ commands"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",  # pylint: disable=redefined-builtin
        columns: str = "default",
        query_str: str = " ",
        sqobj=None,
    ) -> None:
        self.ctxt = context.get_context().ctxt
        self._cfg = self.ctxt.cfg
        self._schemas = self.ctxt.schemas
        self.format = format or "text"
        self.columns = columns.split()
        self._json_print_handler = None
        self._additional_help_vars = {}

        if query_str.count('"') % 2 != 0:
            # This happens because nubia strips off the trailing quote
            # if not followed by a blank
            query_str += '"'
        elif query_str.count("'") % 2 != 0:
            # This happens because nubia strips off the trailing quote
            # if not followed by a blank
            query_str += "'"
        self.query_str = query_str.strip()

        if not isinstance(namespace, str):
            print('namespace must be a space separated list of strings')
            return
        if not isinstance(hostname, str):
            print('hostname must be a space separated list of strings')
            return
        if not isinstance(columns, str):
            print('columns must be a space separated list of strings')
            return

        if not namespace and self.ctxt.namespace:
            self.namespace = self.ctxt.namespace
        else:
            self.namespace = namespace.split()
        if not hostname and self.ctxt.hostname:
            self.hostname = self.ctxt.hostname
        else:
            self.hostname = hostname.split()

        if not start_time and self.ctxt.start_time:
            self.start_time = self.ctxt.start_time
        else:
            self.start_time = start_time

        if not end_time and self.ctxt.end_time:
            self.end_time = self.ctxt.end_time
        else:
            self.end_time = end_time

        self.view = view
        self.lvars = {}         # dict of var name and value for subclass

        if not sqobj:
            raise AttributeError('mandatory parameter sqobj missing')

        self.sqobj = sqobj(context=self.ctxt,
                           engine_name=engine,
                           hostname=self.hostname,
                           start_time=self.start_time,
                           end_time=self.end_time,
                           view=self.view,
                           namespace=self.namespace,
                           columns=self.columns)

    @property
    def cfg(self):
        '''Return the config'''
        return self._cfg

    @property
    def schemas(self):
        '''Return the schemas'''
        return self._schemas

    @ command("help", help="show help for a command")
    @ argument("command", description="command to show help for")
    # pylint: disable=redefined-outer-name
    def help(self, command: str = ''):
        """Show help for a command

        Args:
            command (str, optional): Name of the cmd. Defaults to 'show'.
        """
        if any(x for x in [self.namespace, self.hostname, self.view,
                           self.start_time, self.end_time, self.query_str]):
            print(Fore.RED + "Error: Only accepeted options is command")
            return 1
        if (self.columns != ["default"]) or (self.format != "text"):
            print(Fore.RED + "Error: Only accepted options is command")
            return 1

        if not command:
            mbrs = inspect.getmembers(self)
            verbs = {x[0]: x[1] for x in mbrs
                     if inspect.ismethod(x[1]) and not x[0].startswith('_')}
            print(f"{self.sqobj.table}: " + Fore.CYAN + f"{self.__doc__}" +
                  Style.RESET_ALL)
            verbs = {x: verbs[x] for x in verbs if x != 'get_plugins'}
            print("\nSupported verbs are: ")
            for verb in verbs:
                docstr = inspect.getdoc(verbs[verb])
                if docstr:
                    docstr = docstr.splitlines()[0]
                else:
                    docstr = ''
                verb = verb.replace('aver', 'assert')
                print(f" - {verb}: " + Fore.CYAN + f"{docstr}" +
                      Style.RESET_ALL)
        else:
            self._do_help(self.sqobj.table, command)

        return 0

    def _do_help(self, table: str, verb: str = 'show'):
        """Show help for a command

        Args:
            table (str): the table name
            verb (str, optional): the name of the command. Defaults to 'show'.
        """

        # Determine if the method is present
        if verb == 'assert':
            verb = 'aver'
        fn = inspect.getattr_static(self, verb, None)
        if not fn:
            print(f"No information about {table} {verb}")
            return

        help_dict = inspect.getattr_static(
            self, '__arguments_decorator_specs', None)

        args = list(filter(lambda x: not x.startswith('_') and
                           x not in ['ctxt', 'sqobj', 'lvars'],
                           self.__dict__))
        if self.lvars:
            args += list(self.lvars.keys())

        fnargs = inspect.getfullargspec(fn)
        if fnargs:
            args += list(filter(lambda x: x != 'self', fnargs.args))
            help_dict.update(inspect.getattr_static(
                fn, '__arguments_decorator_specs', {}))

        if not args:
            print(f"No argument information about {table} {verb}")
            return

        helpstr = inspect.getattr_static(self, '__arguments_decorator_specs',
                                         None)
        # Function help string
        docstr = inspect.getdoc(fn)
        if docstr:
            docstr = docstr.splitlines()[0]
        else:
            docstr = ''

        print(f"{table} {verb}: " + Fore.CYAN +
              f"{docstr}" + Style.RESET_ALL)
        if not args:
            print(f"No argument information about {table} {verb}")
            return

        print(Fore.YELLOW + '\nUse quotes when providing more than one value'
              + Style.RESET_ALL)
        print(Fore.YELLOW + '\nArguments:' + Style.RESET_ALL)

        for arg in sorted(args):
            if arg not in help_dict:
                helpstr = 'No description'
            else:
                helpstr = help_dict[arg].description

            print(f" - {arg}: " + colorize(helpstr, Fore.CYAN))

    def _pager_print(self, df: pd.DataFrame) -> None:
        '''To support paging'''

        if isinstance(df, pd.DataFrame) and df.index.dtype == 'int64':
            df.reset_index(drop=True, inplace=True)

        if self.ctxt.pager:
            screen_lines = (shutil.get_terminal_size((80, 20)).lines - 2)
            bufio = StringIO()

            print(df, file=bufio)
            lines = bufio.getvalue().split('\n')
            curline = 0
            total_lines = len(lines)
            while curline < total_lines:
                print('\n'.join(lines[curline:curline+screen_lines]))
                curline += screen_lines
                if curline < total_lines:
                    try:
                        _ = prompt('Hit <ENTER> to continue, <CTRL-D> to quit')
                    except EOFError:
                        break
        else:
            print(df)

    # pylint: disable=too-many-statements
    def _gen_output(self, df: pd.DataFrame, json_orient: str = "records",
                    dont_strip_cols: bool = False, sort: bool = True):

        if ('error' in df.columns or
                ('hopError' in df.columns and not df.empty
                 and (df.hopError != '').all())):
            retcode = 1
            max_colwidth = None
            cols = df.columns.tolist()
            is_error = True
        else:
            max_colwidth = self.ctxt.col_width
            retcode = 0
            if self.columns not in [['default'], ['*']]:
                cols = self.columns
            else:
                cols = df.columns.tolist()
            is_error = False

        max_rows = self.ctxt.max_rows
        all_columns = self.ctxt.all_columns

        if dont_strip_cols or not all(item in df.columns for item in cols):
            cols = df.columns.tolist()

        if self.format == 'json':
            if self._json_print_handler:
                print(df[cols].to_json(
                    default_handler=self._json_print_handler,
                    orient=json_orient))
            else:
                print(df[cols].to_json(orient=json_orient))
        elif self.format == 'csv':
            print(df[cols].to_csv())
        elif self.format == 'markdown':
            print(df[cols].to_markdown())
        elif (self.format == 'devconfig' and
              self.sqobj.table == "devconfig" and
              ('config' in df.columns)):
            for row in df.itertuples():
                self._pager_print(row.config)
        else:
            if 'active' in df.columns:
                df['active'] = np.where(df.active, '+', '-')
            if (not (self.view == "all" or self.columns == ['*'] or
                     (self.start_time and self.end_time) or dont_strip_cols
                     or 'timestamp' in self.columns)):
                if 'timestamp' in cols:
                    cols.remove('timestamp')

            with pd.option_context(
                    'format.precision', 3,
                    'display.width', None,
                    'display.max_columns', None,
                    'display.max_colwidth', max_colwidth,
                    'display.expand_frame_repr', not all_columns,
                    'display.max_rows', max_rows):

                df = self.sqobj.humanize_fields(df)
                if df.empty:
                    print(df[cols])
                elif sort:
                    if is_error:
                        print(df[cols])
                    else:
                        if (((self.start_time and self.end_time) or
                             self.view == "all") and
                                'timestamp' in df.columns):
                            sort_fields = ['timestamp']
                        else:
                            sort_fields = [x for x in self.sqobj.sort_fields
                                           if x in df.columns and x in cols]
                        if sort_fields:
                            self._pager_print(
                                df[cols].sort_values(by=sort_fields,
                                                     key=natsort_keygen()))
                        else:
                            self._pager_print(df[cols])
                else:
                    self._pager_print(df[cols])

        return retcode

    def _assert_gen_output(self, df):
        if df.empty:
            result = 0
        elif df.columns.to_list() == ['error']:
            result = 1
        elif df.loc[df['result'] != "pass"].empty:
            result = 0
        else:
            result = -1

        self._gen_output(df)
        if self.format == 'text':
            if result == 0:
                print("Assert passed")
            else:
                print("Assert failed")
        return result

    def _invoke_sqobj(self, fn, **kwargs) -> pd.DataFrame:

        try:
            df = fn(**kwargs)
        except Exception as ex:  # pylint: disable=broad-except
            if isinstance(ex, UserQueryError):
                df = pd.DataFrame({'error': [f'ERROR: UserQueryError: {ex}']})
            else:
                if self.ctxt.debug:
                    raise ex
                df = pd.DataFrame({'error': [f'ERROR: {ex}']})

        return df

    @command("describe", help="describe the table and its fields")
    def describe(self):
        """Display the schema of the table

        Returns:
            [type]: 0 or error
        """
        now = time.time()

        for x in list(filter(lambda x: not x.startswith('_') and
                             x not in ['ctxt', 'sqobj', 'query_str',
                                       'format', 'columns', 'lvars'],
                             self.__dict__)):
            if getattr(self, x):
                print(Fore.RED + "Error: No arguments allowed for describe")
                return 1

        if any(x for x in self.lvars.values()):
            print(Fore.RED + "Error: No arguments allowed for describe")
            return 1

        df = self._invoke_sqobj(self.sqobj.describe,
                                hostname=self.hostname,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)

    def _print_depracation_command_warning(self,
                                           new_command: str,
                                           sub_command: str,
                                           dep_sub_command: str = None):
        """Print depracation command warning

        Args:
            new_command (str): command
            command (str): _description_
            dep_command (str, optional): _description_. Defaults to None.
        """
        if not dep_sub_command:
            dep_sub_command = sub_command
        dep_command = self.sqobj.table
        warning_msg = deprecated_command_warning(dep_command,
                                                 dep_sub_command,
                                                 new_command,
                                                 sub_command)
        cprint(colored(warning_msg, 'yellow'))

    def _add_result_columns(self, columns: List[str]) -> List[str]:
        """Append the 'result' column to the input columns if necessary.

        This function is used to always retrieve the 'result' column from the
        assertions which is necessary in the _assert_gen_output function.

        The 'result' column is not added if the input columns are '*' or
        'default' because the column will be already present

        Args:
            columns (List[str]): input columns

        Returns:
            List[str]: updated columns
        """
        return (
            columns
            if (columns in [['default'], ['*']] or
                'result' in columns)
            else columns + ['result']
        )


class SqTableCommand(SqCommand):
    """Base Command Class for table commands"""

    @command("summarize", help='produce a summarize of the data')
    def summarize(self, **kwargs):
        """Summarize relevant information about the table"""
        now = time.time()

        summarize_df = self._invoke_sqobj(
            self.sqobj.summarize, namespace=self.namespace,
            hostname=self.hostname, query_str=self.query_str,
            **self.lvars,
            **kwargs
        )

        # for readability syntesize output when dict has more than 3 items
        for index, row in summarize_df.iterrows():
            for col in summarize_df:
                if isinstance(row[col], dict) and len(row[col]) > 3:
                    summarize_df.loc[index, col] = len(row[col])

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(summarize_df, json_orient='columns')

    @command("show")
    def show(self):
        """Show address info
        """
        # Get the default display field names
        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        df = self._invoke_sqobj(self.sqobj.get,
                                hostname=self.hostname,
                                columns=self.columns,
                                query_str=self.query_str,
                                namespace=self.namespace,
                                **self.lvars,
                                )
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)

    @command("unique", help="find the list of unique items in a column")
    @argument("count", description="include count of times a value is seen",
              choices=['True'])
    def unique(self, count='', **kwargs):
        """Get unique values (and counts) associated with requested field"""
        now = time.time()

        df = self._invoke_sqobj(self.sqobj.unique,
                                hostname=self.hostname,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                count=count,
                                **self.lvars,
                                **kwargs,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if 'error' in df.columns or df.empty:
            return self._gen_output(df)

        if not count:
            return self._gen_output(df.sort_values(by=[df.columns[0]]),
                                    dont_strip_cols=True)
        else:
            return self._gen_output(
                df.sort_values(by=['numRows', df.columns[0]]),
                dont_strip_cols=True)

    @ command("top", help="find the top n values for a field")
    @ argument("count", description="number of rows to return")
    @ argument("what", description="numeric field to get top values for")
    @ argument("reverse", description="return bottom n values",
               choices=['True', 'False'])
    def top(self, count: int = 5, what: str = '', reverse: str = 'False',
            **kwargs) -> int:
        """Return the top n values for a field in a table

        Args:
            count (int, optional): Number of entries to return. Defaults to 5
            what (str, optional): Field name to use for largest/smallest val
            reverse (bool, optional): Reverse and return n smallest

        Returns:
            int: 0 or error code
        """
        now = time.time()

        df = self._invoke_sqobj(self.sqobj.top,
                                hostname=self.hostname,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                what=what, count=count,
                                reverse=ast.literal_eval(reverse),
                                **self.lvars,
                                **kwargs,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if 'error' in df.columns:
            return self._gen_output(df)

        if not df.empty:
            df = self.sqobj.humanize_fields(df)
            return self._gen_output(df, dont_strip_cols=True, sort=False)
        else:
            return self._gen_output(df)

    @ command("help", help="show help for a command")
    @ argument("command", description="command to show help for",
               choices=['show', 'unique', 'summarize', 'assert', 'describe',
                        'top', "lpm"])
    # pylint: disable=redefined-outer-name
    def help(self, command: str = ''):
        return super().help(command)



import numpy as np
import pandas as pd

from suzieq.engines.pandas.engineobj import SqPandasEngine


class EvpnvniObj(SqPandasEngine):
    '''Backend class to handle manipulating evpnVni table with pandas'''

    def __init__(self, baseobj):
        super().__init__(baseobj)
        self._assert_result_cols = ['namespace', 'hostname', 'vni', 'type',
                                    'vrf', 'macaddr', 'timestamp', 'result',
                                    'assertReason']

    @staticmethod
    def table_name():
        '''Table name'''
        return 'evpnVni'

    def get(self, **kwargs) -> pd.DataFrame:
        """Class-specific to extract info from addnl tables"""

        columns = kwargs.pop('columns', [])
        query_str = kwargs.pop('query_str', [])

        addnl_fields = []
        fields = self.schema.get_display_fields(columns)
        self._add_active_to_fields(kwargs.get('view', self.iobj.view), fields,
                                   addnl_fields)

        if 'ifname' not in fields+addnl_fields:
            addnl_fields.append('ifname')

        if 'remoteVtepCnt' in fields:
            getVtepCnt = True
            if 'remoteVtepList' not in fields:
                addnl_fields.append('remoteVtepList')
        else:
            getVtepCnt = False

        user_query_cols = self._get_user_query_cols(query_str)
        addnl_fields += [x for x in user_query_cols if x not in addnl_fields]

        df = super().get(addnl_fields=addnl_fields, columns=fields, **kwargs)
        if df.empty:
            return df

        if getVtepCnt:
            insidx = df.columns.tolist().index('remoteVtepList')
            df.insert(insidx+1, 'remoteVtepCnt',
                      df.remoteVtepList.str.len())

        # See if we can retrieve the info to fill out the rest of the data
        # Start with VLAN values
        if 'vlan' not in df.columns and 'vrf' not in df.columns:
            return df.reset_index(drop=True)[fields]

        iflist = [x for x in df[df.vlan == 0]['ifname'].to_list()
                  if x and x != 'None']
        if iflist:
            ifdf = self._get_table_sqobj('interfaces', start_time='').get(
                namespace=kwargs.get('namespace', []), ifname=iflist,
                columns=['namespace', 'hostname', 'ifname', 'state', 'vlan',
                         'vni'])

            if not ifdf.empty:
                df = df.merge(ifdf, on=['namespace', 'hostname', 'ifname'],
                              how='left', suffixes=('', '_y')) \
                    .drop(columns=['timestamp_y', 'state_y', 'vni_y'],
                          errors='ignore')

                df['vlan'] = np.where(
                    df['vlan_y'].notnull(), df['vlan_y'], df['vlan'])
                df = df.drop(columns=['vlan_y']) \
                       .fillna({'vlan': 0}) \
                       .astype({'vlan': int})

        # Now we try to see if we can get the VRF info for all L2 VNIs as well
        ifdf = self._get_table_sqobj('interfaces') \
            .get(namespace=kwargs.get('namespace', []), type=['vlan'],
                 columns=['namespace', 'hostname', 'vlan', 'master']) \
            .query('master != ""') \
            .reset_index(drop=True)
        if not ifdf.empty and not df.empty and 'vrf' in df.columns:
            df = df.merge(ifdf, left_on=['namespace', 'hostname', 'vlan'],
                          right_on=['namespace', 'hostname', 'vlan'],
                          how='outer', suffixes=('', '_y'))

            # We fix the VRF from L2 VNIs from the SVI interface's master info
            # and we drop those entries that are just local VLANs, without VNI
            df['vrf'] = np.where(df['vrf'] == "", df['master'], df['vrf'])
            df = df.drop(columns=['master', 'timestamp_y'], errors='ignore') \
                   .dropna(subset=['vni'])

        # Fill out the numMacs and numArps columns if we can
        if 'numMacs' in df.columns:
            vlanlist = list(set(df[df.numMacs == 0]['vlan'].to_list()))
        else:
            vlanlist = []
        if vlanlist:
            if 'numMacs' in df.columns:
                macdf = self._get_table_sqobj('macs').get(
                    namespace=kwargs.get('namespace'))
                if not macdf.empty:
                    mac_count_df = macdf \
                        .query('flags.isin(["dynamic", "remote"])') \
                        .groupby(['namespace', 'hostname', 'vlan']) \
                        .macaddr.count().reset_index()
                    df = df.merge(mac_count_df,
                                  on=['namespace', 'hostname', 'vlan'],
                                  suffixes=['', '_y'], how='left') \
                        .drop(columns=['numMacs'], errors='ignore') \
                        .rename(columns={'macaddr': 'numMacs'}) \
                        .fillna({'numMacs': 0})

        for x in ['vlan', 'vni']:
            if x in df.columns:
                df[x] = np.where(df[x].isnull(), 0,
                                 df[x].astype(int))

        df = self._handle_user_query_str(df, query_str)
        return df.reset_index(drop=True)[fields]

    def _count_macs(self, x, df):
        return df[(df.namespace == x['namespace']) & (df.vlan == x['vlan']) &
                  (df.hostname == x['hostname'])].count()

    def summarize(self, **kwargs):
        self._init_summarize(**kwargs)
        if self.summary_df.empty:
            return self.summary_df

        self._summarize_on_add_field = [
            ('uniqueVtepCnt', 'priVtepIp', 'nunique'),
            ('uniqueVniCnt', 'vni', 'nunique'),
        ]

        l3vni_count = self.summary_df.query('type == "L3" and vni != 0') \
            .groupby(by=['namespace'])['vni'].count()
        for ns in self.ns.keys():
            if l3vni_count.get(ns, 0):
                self.ns[ns]['mode'] = 'symmetric'
            else:
                self.ns[ns]['mode'] = 'asymmetric'
        self.summary_row_order.append('mode')

        self._summarize_on_add_with_query = [
            ('uniqueL3VniCnt', 'type == "L3" and vni != 0', 'vrf', 'nunique'),
            ('uniqueL2VniCnt', 'type == "L2"', 'vni', 'nunique'),
            ('uniqueMulticastGroups',
             'mcastGroup != "0.0.0.0" and mcastGroup != ""', 'mcastGroup',
             'nunique'),
            ('vnisUsingMulticast',
             'type == "L2" and replicationType == "multicast"',
             'vni', 'nunique'),
            ('vnisUsingIngressRepl',
             'type == "L2" and replicationType == "ingressBGP"',
             'vni', 'nunique')
        ]

        self._gen_summarize_data()

        # To summarize accurately, we need to explode the remoteVteps
        # column from a list to an individual entry for each
        # remoteVteps in that list. The resulting set can be huge if them
        # number of Vteps times the ports is huge.
        #
        # the 'explode' only works post pandas 0.25

        self.summary_df = self.summary_df.explode(
            'remoteVtepList').dropna(how='any')
        self.nsgrp = self.summary_df.groupby(by=["namespace"])

        if not self.summary_df.empty:
            herPerVtepCnt = self.summary_df.groupby(
                by=['namespace', 'hostname'])['remoteVtepList'].nunique()
            self._add_stats_to_summary(herPerVtepCnt, 'remoteVtepsPerVtepStat',
                                       filter_by_ns=True)
        self.summary_row_order.append('remoteVtepsPerVtepStat')

        self._post_summarize(check_empty_col='uniqueVtepCnt')
        return self.ns_df.convert_dtypes()

    def aver(self, **kwargs) -> pd.DataFrame:

        result = kwargs.pop('result', 'all')
        kwargs.pop('columns', None)

        df = self._init_assert_df(**kwargs)

        res_cols = self._assert_result_cols

        if df.empty:
            df = pd.DataFrame(columns=res_cols)
            if result != 'pass':
                df['assertReason'] = 'No data found'
                df['result'] = 'fail'
            return df[res_cols]

        df["assertReason"] = [[] for _ in range(len(df))]

        df['assertReason'] += df.apply(
            lambda x: ['vni down'] if x['state'] != "up" else [],
            axis=1)

        mcast_df = df.query('mcastGroup != "0.0.0.0"')
        if not mcast_df.empty:
            self._validate_vni_replication(df)
            self._validate_mcast_vni_consistency(df, mcast_df)
        else:
            self._validate_all_vteps_known(df)

        self._validate_vni_vrf_consistency(df)

        self._validate_anycast_mac(df)

        # Fill out the assert column
        df['result'] = df.apply(lambda x: 'pass'
                                if len(x.assertReason) == 0 else 'fail',
                                axis=1)

        # Force VNI to be an int
        df['vni'] = np.where(df.vni.isnull(), 0,
                             df.vni.astype(int))

        if result == 'fail':
            df = df.query('assertReason.str.len() != 0')
        elif result == "pass":
            df = df.query('assertReason.str.len() == 0')

        return df[res_cols]

    def _init_assert_df(self, **kwargs):
        '''Prepare for running asserts'''

        addnl_cols = kwargs.pop('addnl_cols', [])
        kwargs.pop("columns", None)  # Loose whatever's passed
        req_cols = ['*']
        if addnl_cols:
            req_cols += [ac for ac in addnl_cols if ac not in req_cols]

        df = self.get(columns=req_cols, **kwargs)
        addr_df = self._get_table_sqobj('address').get(
            namespace=kwargs.get('namespace', []),
            hostname=kwargs.get('hostname', []), type=['vlan'],
            columns=['namespace', 'hostname', 'ifname', 'vlan', 'vrf',
                     'macaddr']) \
            .reset_index(drop=True)
        df = df.merge(addr_df,
                      on=['namespace', 'hostname', 'vlan'],
                      how='left') \
            .drop(columns=['vrf_x'], errors='ignore') \
            .rename(columns={'vrf_y': 'vrf'})

        vni_vrf_df = df.groupby(by=['namespace', 'vni'])['vrf'] \
                       .nunique() \
                       .reset_index() \
                       .rename(columns={'vrf': 'vrfCnt'})
        df = df.merge(vni_vrf_df, on=['namespace', 'vni'], how='left')
        return df

    def _validate_vni_replication(self, df):
        '''A VNI MUST the same replication model on all hosts'''
        repl_df = df.groupby(['namespace', 'vni'])['replicationType'] \
                    .nunique() \
                    .reset_index()

        df = df.merge(repl_df, on=['namespace', 'vni'])
        df['assertReason'] += df.replicationType_y.apply(
            lambda x: ['inconsistent replication across hosts']
            if x != 1 else [])
        df = df.drop(columns=['replicationType_y'])

        return df

    def _validate_mcast_vni_consistency(self, df, mcast_df):
        '''All VNIs have at most one multicast group associated'''

        mismatched_vni_df = mcast_df \
            .groupby(by=['namespace', 'vni'])['mcastGroup'] \
            .unique() \
            .reset_index() \
            .dropna() \
            .query('mcastGroup.str.len() != 1')

        if not mismatched_vni_df.empty:
            df['assertReason'] += df.apply(
                lambda x, err_df: ['VNI has multiple mcast group']
                if (x['namespace'] in err_df['namespace'] and
                    x['vni'] in err_df['vni']) else [], axis=1,
                args=(mismatched_vni_df, ))

    def _validate_all_vteps_known(self, df):
        '''Every VTEP MUST about every other VTEP for a given VNI with HER'''
        df['pickVtepIp'] = np.where(df.secVtepIp == "", df.priVtepIp,
                                    df.secVtepIp)
        df['avt'] = df.apply(
            lambda x: np.sort(np.append(x['remoteVtepList'],
                                        [x['pickVtepIp']])),
            axis=1)
        # pylint: disable=unnecessary-lambda
        df['allVteps'] = df.avt.apply(lambda x: ','.join(x))
        known_vteps = df.groupby(by=['namespace', 'vni'])['allVteps'] \
                        .nunique() \
                        .reset_index()
        df = df.merge(known_vteps, on=['namespace', 'vni']) \
               .drop(columns=['avt', 'allVteps_x', 'pickVtepIp']) \
               .rename(columns={'allVteps_y': 'allVtepsCnt'})
        df['assertReason'] += df.allVtepsCnt.apply(
            lambda x: ['Some VTEPs are unknown to all'] if x != 1 else [])

    def _validate_vni_vrf_consistency(self, df):
        '''A L2 VNI MUST be mapped to the same L3VNI on every host'''

        df['assertReason'] += df.vrfCnt.apply(
            lambda x: ['Inconsistent VNI-VRF mapping'] if (x > 1) else [])

    def _validate_anycast_mac(self, df):
        '''All non-L2-only segments MUST have the same anycast MACaddr'''
        rest_df = df.query('type == "L2" and macaddr != "00:00:00:00:00:00"')
        if rest_df.macaddr.nunique() != 1:
            df['assertReason'] += df.vrfCnt.apply(
                lambda x: ['anycast MAC not unique across VNI']
                if x else [])

    def _validate_frr_her(self, df, **kwargs):
        '''In FRR, with ingress replication, look for HER list consistency'''

        mac_df = self._get_table_sqobj('macs') \
            .get(namespace=kwargs.get("namespace", ""),
                 macaddr=["00:00:00:00:00:00"], remoteVtepIp=['any'])

        # # Assert that we have HER for every remote VTEP; Cumulus/SONiC
        if not mac_df.empty:
            df['assertReason'] += df.apply(self._is_her_good,
                                           args=(mac_df, ), axis=1)

    def _is_her_good(self, row, mac_df):
        reason = []

        if row['type'] == 'L3':
            return reason

        her_list = mac_df[(mac_df['hostname'] == row['hostname']) &
                          (mac_df['namespace'] == row['namespace']) &
                          (mac_df['oif'] == row['ifname'])]['remoteVtepIp'] \
            .tolist()
        missing_hers = set(row['remoteVtepList']).difference(set(her_list))
        if not missing_hers:
            return reason
        return [f'HER is missing VTEPs {missing_hers}']



from typing import List
from ipaddress import ip_address, ip_network

import pandas as pd

from suzieq.engines.pandas.engineobj import SqPandasEngine
from suzieq.shared.utils import convert_macaddr_format_to_colon


class NetworkObj(SqPandasEngine):
    '''Backend class to handle network queries'''

    @staticmethod
    def table_name():
        '''Table name'''
        return 'network'

    def find(self, **kwargs):
        '''Find the information requsted:

        address: given a MAC or IP address, find the first hop switch its
                 connected to
        '''

        addrlist = kwargs.pop('address', [])
        kwargs.pop('columns', ['default'])
        query_str = kwargs.pop('query_str', '')
        fields = self.schema.get_display_fields(['default'])

        if ((self.iobj.start_time and self.iobj.end_time) or
                (self.iobj.view == 'all')):
            fields.insert(0, 'active')
            fields.append('timestamp')
        dflist = []
        if isinstance(addrlist, str):
            addrlist = [addrlist]
        for addr in addrlist:
            df = self._find_address(addr, **kwargs)
            if not df.empty:
                dflist.append(df)
        if dflist:
            df = pd.concat(dflist)

        if df.empty:
            return df

        if not df.empty and query_str:
            return df.query(query_str)

        return df[fields]

    def _find_asn(self, asn: List[str], **kwargs) -> pd.DataFrame:
        """Find the hosts with the ASN listed

        Args:
            asn (List[str]): List of ASNs, space separated

        Returns:
            pd.DataFrame: The dataframe with the info. The columns are-
                          namespace, hostname, asn
        """

        if not asn:
            return pd.DataFrame()

        _ = kwargs.pop('vlan', [])
        try:
            asn = [int(x) for x in asn]
        except ValueError:
            return pd.DataFrame({'error': ['ASNs must be an integers']})

        return self._get_table_sqobj('bgp') \
            .get(asn=asn, columns=['default'], **kwargs)

    def _find_address(self, addr: str, **kwargs) -> pd.DataFrame:
        """Find the origin or the first hop network device that owns this addr

        Args:
            addr (str): address to search for

        Returns:
            pd.DataFrame: Dataframe with the relevant information
        """

        vlan = kwargs.pop('vlan', '')
        vrf = kwargs.pop('vrf', '')
        columns = kwargs.get('columns', ['default'])
        cols = self.schema.get_display_fields(columns)

        if ((self.iobj.start_time and self.iobj.end_time) or
                (self.iobj.view == 'all')):
            cols.insert(0, 'active')
            cols.append('timestamp')
        # Convert Cisco-style MAC address to standard MAC addr format,
        # and lowercase all letters of the alphabet
        addr = convert_macaddr_format_to_colon(addr)

        addr_df = self._find_addr_arp(addr, **kwargs)
        if addr_df.empty:
            # Is this a locally attached interface IP to a device we're polling
            df = self._get_table_sqobj('address') \
                .get(vrf=vrf, address=[addr],
                     columns=['namespace', 'hostname', 'ifname', 'vlan', 'vrf',
                              'ipAddress', 'macaddr', 'timestamp'],
                     **kwargs)
            if df.empty:
                return addr_df

            # Only pick the entry with the address specified
            df = df.explode('ipAddress')
            df['ipAddress'] = df.ipAddress.str.split('/').str[0]
            if any(x in addr for x in ['::', '.']):
                df = df.query(f'ipAddress == "{addr}"').reset_index(drop=True)
            else:
                # If we're querying by MAC Addr and an interface has no IP,
                # without fillna, the entries are dropped due to dropna below
                # Arista's interfaces return a 0.0.0.0 if the primary address
                # isn't set. Weed out these entries

                df = df.query(f'macaddr == "{addr}"') \
                       .fillna('') \
                       .query('ipAddress != "0.0.0.0"') \
                       .reset_index(drop=True)

            df['type'] = 'interface'
            df['l2miss'] = False
            addr_df = df

        addr_df = self._find_first_hop_attach(addr_df)
        if vlan and not addr_df.empty:
            addr_df = addr_df.query(f'vlan == {vlan}')
        if vrf and not addr_df.empty:
            addr_df = addr_df.query(f'vrf == "{vrf}"')

        if addr_df.empty:
            return addr_df

        # If routed interface and unnumbered interfaces are used, we need to
        # identify the primary interface
        addr_df = addr_df \
            .dropna() \
            .drop_duplicates() \
            .reset_index(drop=True)

        if addr_df.type.unique().tolist() == ['routed']:
            addr_df = self._add_primary_interface(addr_df, addr)

        return addr_df[cols]

    def _find_bond_members(self, namespace: str, hostname: str,
                           ifname: str, timestamp: str) -> List[str]:
        """This function returns the member ports of a bond interface
        Args:
            namespace (str): Namespace of the host
            hostname (str): Hostname of the host
            ifname (str): Interface name of the bond
            timestamp (str): time string, of str(timestamp) format

        Returns:
            mbr_list(List[str]): List of member ports or input ifname
        """

        # get list  of namespaces, hostnames and ifnames to get info for

        ifdf = self._get_table_sqobj('interfaces',
                                     start_time='',
                                     end_time=timestamp) \
            .get(namespace=[namespace], hostname=[hostname], master=[ifname])

        if not ifdf.empty:
            return ifdf.ifname.unique().tolist()

        return [ifname]

    # pylint: disable=too-many-statements
    def _find_addr_arp(self, addr: str, **kwargs) -> pd.DataFrame:
        """Find the origin or the first hop network device that owns this addr

        Args:
            addr (str): address to search for

        Returns:
            pd.DataFrame: Dataframe with the relevant information
        """

        if any(x in addr for x in ['::', '.']):
            arpdf = self._get_table_sqobj('arpnd').get(
                ipAddress=addr.split(), **kwargs)
        else:
            arpdf = self._get_table_sqobj('arpnd').get(
                macaddr=addr.split(), **kwargs)

        # This var tracks the set of hosts that were active and had
        # local ARP entries when last seen. Its used when view=all or
        # start and end times are specified.
        active_local_hosts = set()
        result = []
        if arpdf.empty:
            return pd.DataFrame(result)

        # When providing view=all or a start-time/end-time, since we're
        # using data from multiple table to produce the final output, we
        # need to know how to handle the time when we look for data in those
        # additional tables. For example, consider an ARP entry exists at time
        # t1, and then has changes at t2 and t3. Now lets say the user wants
        # to see all changes or changes between t1 and t3. WHen processing
        # the entries for row t1, We need to ensure that we look for data in
        # other tables only before t2, only before t3 when processing the
        # entry at time t2 and so on. We're figuring out the time window
        # with the code below.
        if not arpdf.empty:
            key_cols = ['namespace', 'hostname', 'ipAddress']
            arpdf = arpdf.sort_values(by=key_cols + ['timestamp'])
            arpdf['next_ts'] = arpdf.groupby(by=key_cols)['timestamp'] \
                                    .shift(-1)

        namespaces = arpdf.namespace.unique().tolist()
        hostnames = arpdf.hostname.unique().tolist()
        # Weed out MACVLAN interfaces on Cumulus that are for VIPs
        ifaces = [x for x in arpdf.oif.unique().tolist()
                  if not x.endswith('-v0')]
        macaddrs = arpdf.macaddr.unique().tolist()

        ifdf = self._get_table_sqobj('interfaces', start_time='') \
            .get(namespace=namespaces, hostname=hostnames,
                 ifname=ifaces)
        # We don't explicitly look for only local-only entries because
        # that precludes us addresses that were not remote in the past.
        macdf = self._get_table_sqobj('macs', start_time='') \
                    .get(namespace=namespaces, hostname=hostnames,
                         macaddr=macaddrs,
                         columns=['namespace', 'hostname', 'vlan', 'macaddr',
                                  'oif', 'remoteVtepIp', 'flags', 'active',
                                  'timestamp'])

        for arp_row in arpdf.itertuples():
            # We ignore the start time because the entries such as interface
            # may not have changeed in a very long time. Using the
            # start-time if one is provided can make us not return the
            # relevant data in situations where say, the interface has been
            # stable long before.
            # the start time.
            row_end_time = arp_row.next_ts or self.iobj.end_time
            tmpres = {}

            if getattr(arp_row, 'error', None):
                continue
            # Remote ARP entries are not what we need to track down locally
            # attached addresses. The if removes any always remote entries
            # from being considered.
            if (arp_row.remote and
                    arp_row.hostname not in active_local_hosts):
                continue

            if arp_row.oif.endswith('-v0'):
                # Ignore VRR interfaces in Cumulus, we got them via the
                # normal interface entry (without -v0)
                continue
            oif = arp_row.oif

            active = getattr(arp_row, 'active', True)
            # If an entry goes from local to being deleted, remove it
            if not active and arp_row.hostname in active_local_hosts:
                active_local_hosts.remove(arp_row.hostname)

            if active and not arp_row.remote:
                active_local_hosts.add(arp_row.hostname)

            row_ifdf = ifdf.query(f'namespace=="{arp_row.namespace}" and '
                                  f'hostname=="{arp_row.hostname}" and '
                                  f'ifname=="{arp_row.oif}"')
            # We need to select the entries that are closest to this
            # arpdf entry's timestamp
            if (not row_ifdf.empty and
                (self.iobj.end_time or self.iobj.view == "all") and
                    row_end_time is not pd.NaT):
                row_ifdf = row_ifdf[row_ifdf['timestamp'] <= row_end_time]

            if not row_ifdf.empty:
                netaddr = ip_address(arp_row.ipAddress)
                row_vrf = row_ifdf.master.unique().tolist()[0] or 'default'
                # The code below checks if the provided address belongs to
                # the interface subnet. If it does, its a bridged address
                # else its a routed address i.e. you use routing to reach it.
                if not ifdf.ipAddressList.apply(
                        lambda subnets, netaddr:
                        False if not subnets.any() else
                        any(netaddr in ip_network(subnet, strict=False)
                            for subnet in subnets),
                        args=(netaddr,)).any():

                    # Routed interface
                    tmpres.update({
                        'active': active,
                        'namespace': arp_row.namespace,
                        'hostname': arp_row.hostname,
                        'vrf': row_vrf,
                        'ipAddress': arp_row.ipAddress,
                        'vlan': row_ifdf.vlan.unique().tolist()[0],
                        'macaddr': arp_row.macaddr,
                        'ifname': oif,
                        'type': 'routed',
                        'l2miss': False,
                        'timestamp': arp_row.timestamp
                    })
                    result.append(tmpres)
                    continue

                # At this point, we're dealing with bridged addresses, and as
                # an endpoint I'm expecting an SVI. Its possible its an
                # interface address on a routed link, but we cover that above
                # or via fetching the address table in the code that calls this
                # function.
                if 'vlan' not in ifdf.type.unique():
                    continue

                # If the entry switched from local to remote, mark it as
                # deleted
                if active and arp_row.remote:
                    # if the entry didnt switch from local to remote, we'd
                    # have continued above
                    # Routed interface
                    tmpres.update({
                        'active': False,
                        'namespace': arp_row.namespace,
                        'hostname': arp_row.hostname,
                        'vrf': row_vrf,
                        'ipAddress': arp_row.ipAddress,
                        'vlan': row_ifdf.vlan.unique().tolist()[0],
                        'macaddr': arp_row.macaddr,
                        'ifname': oif,
                        'type': 'bridged',
                        'l2miss': False,
                        'timestamp': arp_row.timestamp
                    })
                    result.append(tmpres)
                    active_local_hosts.remove(arp_row.hostname)
                    continue

                row_macdf = macdf.query(
                    f'namespace=="{arp_row.namespace}" and '
                    f'hostname=="{arp_row.hostname}" and '
                    f'macaddr=="{arp_row.macaddr}" and '
                    f'vlan=={ifdf.vlan.unique().tolist()}')
                # We need to select the entries that are closest to this
                # arpdf entry's timestamp. We can get duplicate entries
                # due to the various possibilities of mac table updates
                # (the mac table may not have changed in a long time or
                # it might've in the time window of this entry). We use
                # drop_duplicates later to remove duplicate entries.
                if (not row_macdf.empty and
                    (self.iobj.end_time or self.iobj.view == "all") and
                        row_end_time is not pd.NaT):
                    row_macdf = row_macdf[(row_macdf['timestamp']
                                           <= row_end_time) &
                                          (row_macdf['flags'] != "remote")]

                for mac_row in row_macdf.itertuples():
                    tmpres = {}
                    if (not mac_row.oif or mac_row.oif == "vPC Peer-Link" or
                            mac_row.flags == 'remote'):
                        continue
                    oifs = mac_row.oif

                    if not mac_row.active:
                        tmpres.update({
                            'active': active,
                            'namespace': arp_row.namespace,
                            'hostname': arp_row.hostname,
                            'vrf': row_vrf,
                            'ipAddress': arp_row.ipAddress,
                            'vlan': mac_row.vlan,
                            'macaddr': arp_row.macaddr,
                            'ifname': arp_row.oif,
                            'type': 'bridged',
                            'l2miss': True,
                            'timestamp': mac_row.timestamp
                        })
                    else:
                        tmpres.update({
                            'active': active,
                            'namespace': arp_row.namespace,
                            'hostname': arp_row.hostname,
                            'vrf': row_vrf,
                            'ipAddress': arp_row.ipAddress,
                            'vlan': mac_row.vlan,
                            'macaddr': arp_row.macaddr,
                            'ifname': oifs,
                            'type': 'bridged',
                            'l2miss': False,
                            'timestamp': arp_row.timestamp
                        })
                    result.append(tmpres)

                if row_macdf.empty:
                    for ele_vlan in row_ifdf.vlan.unique().tolist():
                        tmpres.update({
                            'active': active,
                            'namespace': arp_row.namespace,
                            'hostname': arp_row.hostname,
                            'vrf': row_vrf,
                            'ipAddress': arp_row.ipAddress,
                            'vlan': ele_vlan,
                            'macaddr': arp_row.macaddr,
                            'ifname':
                            ' '.join(row_ifdf.query(f'vlan == {ele_vlan}')
                                     .ifname.unique().tolist()),
                            'type': 'bridged',
                            'l2miss': True,
                            'timestamp': arp_row.timestamp
                        })
                    result.append(tmpres)

        return pd.DataFrame(result)

    # pylint: disable=too-many-statements
    def _find_first_hop_attach(self, addr_df: pd.DataFrame) -> pd.DataFrame:
        """Find the first hop switch attachment point for this address

        Args:
            addr_df (pd.DataFrame): Dataframe with the address information

        Returns:
            pd.DataFrame: Dataframe with the L2 attachment point information
        """

        # We'll need to resolve bond interfaces first
        l2_addr_df = addr_df
        if l2_addr_df.empty:
            return addr_df

        result = []

        # Calculate the next timestamp window for use with pulling in
        # correct info
        key_cols = ['namespace', 'hostname', 'vrf', 'ipAddress']
        l2_addr_df = l2_addr_df.sort_values(by=key_cols + ['timestamp'])
        l2_addr_df['next_ts'] = l2_addr_df.groupby(by=key_cols)['timestamp'] \
                                          .shift(-1)

        for row in l2_addr_df.itertuples():
            tmpres = {}

            active = getattr(row, 'active', True)
            match_ifname = row.ifname.split('.')[0]
            match_hostname = row.hostname
            match_namespace = row.namespace
            match_vrf = row.vrf
            # We ignore the start time because the entries such as interface
            # may not have changeed in a very long time. Using the
            # start-time if one is provided can make us not return the
            # relevant data in situations where say, the interface has been
            # stable long before.
            if self.iobj.end_time:
                if row.next_ts is pd.NaT:
                    match_endtime = self.iobj.end_time
                else:
                    match_endtime = str(row.next_ts)
            else:
                match_endtime = ''
            macobj = self._get_table_sqobj('macs', start_time='',
                                           end_time=match_endtime)
            lldpobj = self._get_table_sqobj('lldp', start_time='',
                                            end_time=match_endtime)

            while True:
                # Do not attempt to replace tmp_df with l2_addr_df and get
                # rid of do_contiue. For some weird reason, with pandas ver
                # 1.2.5, the explode line below hangs if we just use
                # l2_addr_df

                mbr_ports = self._find_bond_members(
                    match_namespace, match_hostname, match_ifname,
                    match_endtime)

                # If what we have is an address of an interface, don't chase
                # down the rabbit hole for first attach point
                if row.type == 'interface':
                    if match_ifname in mbr_ports:
                        mbr_ports = ''
                    match_ifname = row.ifname
                    break

                lldp_df = lldpobj.get(namespace=[row.namespace],
                                      hostname=[row.hostname],
                                      ifname=mbr_ports)

                if not lldp_df.empty:
                    peer_host = lldp_df.peerHostname.unique().tolist()[0]
                    if row.type == 'routed':
                        match_hostname = peer_host
                        match_ifname = lldp_df.peerIfname.unique().tolist()[0]
                        mbr_ports = self._find_bond_members(
                            match_namespace, match_hostname, match_ifname,
                            match_endtime)
                        # Need to get VRF for the interface
                        ifdf = self._get_table_sqobj('interfaces') \
                            .get(namespace=[match_namespace],
                                 hostname=[match_hostname],
                                 ifname=[match_ifname])
                        if not ifdf.empty:
                            match_vrf = ifdf.master.unique().tolist()[
                                0] or 'default'
                        # Routed ports never go more than one hop
                        break

                    macdf = macobj.get(namespace=[match_namespace],
                                       hostname=[peer_host],
                                       macaddr=row.macaddr,
                                       vlan=str(row.vlan),
                                       columns=['namespace', 'hostname',
                                                'vlan', 'macaddr', 'oif'])
                    if not macdf.empty:
                        match_hostname = peer_host
                        match_ifname = macdf.oif.unique().tolist()[0]
                    else:
                        break
                else:
                    break

            if match_ifname in mbr_ports:
                mbr_ports = ''

            tmpres.update({
                'active': active,
                'namespace': row.namespace,
                'hostname': match_hostname,
                'vrf': match_vrf,
                'ipAddress': row.ipAddress,
                'vlan': row.vlan,
                'macaddr': row.macaddr,
                'ifname': match_ifname,
                'bondMembers': ', '.join(mbr_ports),
                'type': row.type,
                'l2miss': row.l2miss,
                'timestamp': row.timestamp
            })
            result.append(tmpres)

        if result:
            return pd.DataFrame(result)

        return addr_df

    def _add_primary_interface(self, addr_df: pd.DataFrame,
                               addr: str) -> pd.DataFrame:
        """Add the primary interface for this address if unnumbered

        Args:
            addr_df (pd.DataFrame): Dataframe with the address information
            addr (str): address to search for

        Returns:
            pd.DataFrame: Dataframe with primary interface rows added
        """

        hostnsgrp = addr_df.groupby(['hostname', 'namespace', 'vrf'])
        if len(hostnsgrp) > 1:
            # not a set of duplicated interfaces, return
            return addr_df

        # We need to add the empty string to handle junos devices which have
        # the IP address on .0 interfaces which the poller tended to classify
        # as subinterface.
        vrf = [addr_df.vrf[0], '']
        # We have a set of duplicated interfaces, find the primary
        df = self._get_table_sqobj('address', start_time='') \
            .get(namespace=[addr_df.namespace[0]],
                 hostname=[addr_df.hostname[0]],
                 vrf=vrf, type='loopback')

        if df.empty or ((df.ipAddressList.str.len() == 0).all() and
                        (df.ip6AddressList.str.len() == 0).all()):
            # Junos interfaces have IP address on the subinterface, and
            # so look for that.
            ifnames = [f'{x}.0' for x in df.ifname.unique()]
            df = self._get_table_sqobj('address') \
                     .get(namespace=[addr_df.namespace[0]],
                          hostname=[addr_df.hostname[0]],
                          vrf=vrf, ifname=ifnames)

        df = df.explode('ipAddressList').explode('ip6AddressList') \
            .query(f'ipAddressList.str.startswith("{addr}/") or '
                   f'ip6AddressList.str.startswith("{addr}/")') \
            .fillna('') \
            .reset_index(drop=True)

        if not df.empty:
            df['macaddr'] = "00:00:00:00:00:00"
            df['vlan'] = 0
            df['bondMembers'] = ''
            df['type'] = 'interface'
            df['l2miss'] = False
            if '::' in addr:
                df['ipAddress'] = df.ip6AddressList.apply(lambda x:
                                                          x.split('/')[0])
            else:
                df['ipAddress'] = df.ipAddressList.apply(lambda x:
                                                         x.split('/')[0])

            addr_df = addr_df.append(df[['namespace', 'hostname', 'ifname',
                                         'ipAddress', 'vrf', 'vlan', 'type',
                                        'l2miss', 'macaddr', 'bondMembers']])
        return addr_df
